import logging

from pecan import expose, request
from pecan.ext.notario import validate
from uuid import uuid4

from ceph_installer.controllers import error
from ceph_installer.tasks import call_ansible
from ceph_installer import schemas
from ceph_installer import models
from ceph_installer import util


logger = logging.getLogger(__name__)


class AgentController(object):

    @expose(generic=True, template='json')
    def index(self):
        error(405)

    @index.when(method='POST', template='json')
    @validate(schemas.agent_install_schema, handler="/errors/schema")
    def install(self):
        master = request.json.get('master', request.server_name)
        logger.info('defining "%s" as the master host for the minion configuration', master)
        hosts = request.json.get('hosts')
        verbose_ansible = request.json.get('verbose', False)
        extra_vars = util.get_install_extra_vars(request.json)
        extra_vars['agent_master_host'] = master
        identifier = str(uuid4())
        task = models.Task(
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(extra_vars=extra_vars, verbose=verbose_ansible)
        call_ansible.apply_async(
            args=([('agents', hosts)], identifier),
            kwargs=kwargs,
        )

        return task



import logging

from pecan import expose, request
from pecan.ext.notario import validate
from uuid import uuid4

from ceph_installer.controllers import error
from ceph_installer.tasks import call_ansible
from ceph_installer import schemas
from ceph_installer import models
from ceph_installer import util


logger = logging.getLogger(__name__)


class MONController(object):

    @expose('json')
    def index(self):
        # TODO: allow some autodiscovery here so that clients can see what is
        # available
        return dict()

    @expose(generic=True, template='json')
    def install(self):
        error(405)

    @install.when(method='POST', template='json')
    @validate(schemas.mon_install_schema, handler="/errors/schema")
    def install_post(self):
        hosts = request.json.get('hosts')
        install_calamari = request.json.get('calamari', False)
        verbose_ansible = request.json.get('verbose', False)
        extra_vars = util.get_install_extra_vars(request.json)
        extra_vars['calamari'] = install_calamari
        identifier = str(uuid4())
        task = models.Task(
            request=request,
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(
            extra_vars=extra_vars,
            tags="package-install",
            verbose=verbose_ansible,
        )

        call_ansible.apply_async(
            args=([('mons', hosts)], identifier),
            kwargs=kwargs,
        )

        return task

    @expose(generic=True, template='json')
    def configure(self):
        error(405)

    @configure.when(method='POST', template='json')
    @validate(schemas.mon_configure_schema, handler="/errors/schema")
    def configure_post(self):
        monitor_mapping = dict(host=request.json['host'])

        # Only add interface and address if they exist
        for key in ['interface', 'address']:
            try:
                monitor_mapping[key] = request.json[key]
            except KeyError:
                pass
        hosts = util.parse_monitors([monitor_mapping])
        verbose_ansible = request.json.get('verbose', False)
        monitors = request.json.get("monitors", [])
        monitors = util.validate_monitors(monitors, request.json["host"])
        # even with configuring we need to tell ceph-ansible
        # if we're working with upstream ceph or red hat ceph storage
        extra_vars = util.get_install_extra_vars(request.json)
        # this update will take everything in the ``request.json`` body and
        # just pass it in as extra-vars. That is the reason why optional values
        # like "calamari" are not looked up explicitly. If they are passed in
        # they will be used.
        extra_vars.update(request.json)
        if 'verbose' in extra_vars:
            del extra_vars['verbose']
        if 'conf' in extra_vars:
            extra_vars['ceph_conf_overrides'] = request.json['conf']
            del extra_vars['conf']
        if monitors:
            hosts.extend(util.parse_monitors(monitors))
            del extra_vars['monitors']
        if "cluster_name" in extra_vars:
            extra_vars["cluster"] = extra_vars["cluster_name"]
            del extra_vars["cluster_name"]
        del extra_vars['host']
        extra_vars.pop('interface', None)
        extra_vars.pop('address', None)
        identifier = str(uuid4())
        task = models.Task(
            request=request,
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(
            extra_vars=extra_vars,
            skip_tags="package-install",
            verbose=verbose_ansible,
        )
        call_ansible.apply_async(
            args=([('mons', hosts)], identifier),
            kwargs=kwargs,
        )

        return task



import errno
import os
import pecan
import tempfile
import logging
from StringIO import StringIO
from ceph_installer import templates


logger = logging.getLogger(__name__)


def generate_inventory_file(inventory, task_uuid, tmp_dir=None):
    """
    Generates a host file to use with an ansible-playbook call.

    The first argument is a list of tuples that contain the group name as
    the first item in the tuple and a list of hostnames in the second.

    For example:

        [('mons', ['mon.host']), ('osds', ['osd1.host', 'osd1.host'])]
    """
    result = []
    for section in inventory:
        group_name = section[0]
        hosts = section[1]
        result.append("[{0}]".format(group_name))
        if not isinstance(hosts, list):
            hosts = [hosts]
        result.extend(hosts)
    result_str = "\n".join(result) + "\n"
    # if not None the NamedTemporaryFile will be created in the given directory
    tempfile.tempdir = tmp_dir
    inventory_file = tempfile.NamedTemporaryFile(prefix="{0}_".format(task_uuid), delete=False)
    inventory_file.write(result_str)
    inventory_file.close()
    return inventory_file.name


def validate_monitors(monitors, host):
    """
    Ensures that the given host is not included in
    the monitors list. This fixes an issue with users
    trying to provide the monitor they are currently
    configuring in the list of 'monitors'.
    """
    result = []
    for mon in monitors:
        if mon["host"] == host:
            logger.warning("When configuring a mon it can not exist in the list of current monitors.")
            logger.warning("%s removed from monitors.", host)
            continue
        result.append(mon)
    return result


def parse_monitors(monitors):
    """
    Given a list of dictionaries, returns a list of hosts that
    can be used in an ansible inventory. These host lines can include
    host variables as well.

    For example, monitors in this format::

        [
            {"host": "mon0.host", "interface": "eth0"},
            {"host": "mon1.host", "interface": "enp0s8"},
        ]

    Would return the following::

        ["mon0.host monitor_interface=eth0", "mon1.host monitor_interface=enp0s8"]

    Because the API allows for both ``interface`` or ``address`` this utility
    will look for both.  Ideally, only one should be defined, but it is up to
    the client to ensure that the one that is needed is passed to the API.
    """
    hosts = []

    for mon in monitors:
        host = []
        host.append(mon["host"])

        # This is an 'either or' situation. The schema engine does not allow us
        # to play with situations that might have one key or the other. That
        # problem gets solved here by trying to use monitor_interface first and
        # then falling back to its address if defined.
        try:
            host.append("monitor_interface=%s" % mon['interface'])
        except KeyError:
            try:
                host.append("monitor_address=%s" % mon['address'])
            except KeyError:
                # do not append monitor_* and just use the host
                pass

        hosts.append(" ".join(host))

    return hosts


def mkdir(path, exist_ok=True, mode=0755):
    """
    Create a directory if it already exists do not error. Anything else should
    be treated as an error condition.
    """
    try:
        os.makedirs(path, mode=mode)
    except OSError as exc:  # Python >2.5
        if not exist_ok:
            raise
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def api_endpoint(endpoint=None, *args):
    """
    Puts together the API url for ceph_installer, so that we can talk to it. Optionally, the endpoint
    argument allows to return the correct url for specific needs. For example, to create a new task
    with a distinct identifier::

        >>> api_endpoint('tasks', str(uuid()))
        >>> "http://0.0.0.0/api/tasks/47e9a7bc-e27a-4cb3-ab7d-34c54381323f/"

    """
    server = pecan.conf['server']['host']
    port = pecan.conf['server']['port']
    base = "http://%s:%s/api" % (server, port)
    endpoints = {
        'tasks': '%s/tasks/' % base,
        'mon': '%s/mon/' % base,
        'rgw': '%s/rgw/' % base,
        'calamari': '%s/calamari/' % base,
        'osd': '%s/osd/' % base,
        'agent': '%s/agent/' % base,
    }
    url = base
    if endpoint:
        url = endpoints[endpoint]

    if args:
        for part in args:
            url = os.path.join(url, part)
    return url


def which(executable):
    """
    Locate in known system $PATH locations where an executable might exist.
    This will mean that non-system $PATH locations (like virtualenv paths) will
    be fully ignored.
    """
    locations = (
        '/usr/local/bin',
        '/bin',
        '/usr/bin',
        '/usr/local/sbin',
        '/usr/sbin',
        '/sbin',
    )

    for location in locations:
        executable_path = os.path.join(location, executable)
        if os.path.exists(executable_path):
            return executable_path


def get_ceph_ansible_path():
    """
    Try to determine where does the ceph-ansible repo lives. It does this by
    looking into an environment variable first which takes precedence (for
    now).
    """
    try:
        repo_path = os.environ['CEPH_ANSIBLE_PATH']
    except KeyError:
        # TODO: Fallback nicely into looking maybe in some directory that is
        # included with the ceph_installer application or a well know path defined by
        # the packaging of said playbooks
        logger.warning('"CEPH_ANSIBLE_PATH" environment variable is not defined')
        repo_path = '/usr/share/ceph-ansible'

    return repo_path


def get_endpoint(request_url, *args):
    """
    Given a request URL, attempt to determine the full address to the `/setup/`
    endpoint so that we can re-use this when crafting the script.

    This service might not be reached at the same address as the configuration
    states so we try to match the request to ensure a subsequent request that
    can capture stdout/stderr will be correct.

    The ``request_url`` argument is required, and used to infer the base url
    this service was accessed from. All args would then be used to append to
    the main url::

        >>> get_endpoint('http://localhost:8181/some/endpoint', 'api', 'setup')
        'http://localhost:8181/api/setup/'

    ``request_url``: A string representing the full URL like ``http://api.example.com/api/``.
    """
    url = '/'.join(request_url.split('/')[:3])
    if args:
        for part in args:
            url = os.path.join(url, part)
    if not url.endswith('/'):
        return "%s/" % url
    return url


def make_agent_script(url, target_host):
    """
    Create the agent setup script. This is very similar to the
    ``make_setup_script`` with the difference being a trigger to install and
    configure the agent after the provisioning user has been created.
    """
    template = templates.setup_script + templates.agent_script
    ssh_key_address = get_endpoint(url, 'setup', 'key')
    agent_endpoint = get_endpoint(url, 'api', 'agent')
    script = StringIO()
    script.write(
        template.format(
            ssh_key_address=ssh_key_address,
            target_host=target_host,
            agent_endpoint=agent_endpoint,
        )
    )
    script.seek(0)
    return script


def make_setup_script(url):
    """
    Create a setup script. Done dynamically due to the need of identifying the
    correct url used to request this service. The scrip will use this URL to
    properly create the right location for the serving of the public ssh key.
    """
    ssh_key_address = get_endpoint(url, 'setup', 'key')
    script = StringIO()
    script.write(
        templates.setup_script.format(
            ssh_key_address=ssh_key_address,
        )
    )
    script.seek(0)
    return script


def get_install_extra_vars(json):
    """
    Given a request.json dictionary from an install endpoint, this
    method will generate and return a dict to be used as extra_vars
    for ceph_installer.tasks.install.
    """
    extra_vars = dict()
    redhat_storage = json.get("redhat_storage", False)
    logger.info("Setting redhat_storage to %s", redhat_storage)
    use_cdn = json.get("redhat_use_cdn", True)
    logger.info("Setting redhat_use_cdn to %s", use_cdn)
    if redhat_storage:
        extra_vars["ceph_stable_rh_storage"] = True
        if not use_cdn:
            # this tells ceph-ansible to install from the
            # existing repos on the system
            extra_vars["ceph_origin"] = "distro"
        else:
            extra_vars["ceph_stable_rh_storage_cdn_install"] = True
    else:
        # use the latest upstream stable version
        extra_vars["ceph_stable"] = True
    # fetch_directory must be a writable location.
    fetch = os.path.join(os.environ['HOME'], 'fetch')
    extra_vars["fetch_directory"] = fetch

    return extra_vars


def get_osd_configure_extra_vars(json):
    """
    Given a request.json dictionary from the api/osd/configure/ endpoint,
    this method will generate and return a dict to be used as extra_vars
    for ceph_installer.tasks.install.

    This new dictionary "translates" exposed/documented keys and values to
    variables that ceph-ansible can consume correctly. That is the reason why
    some of these keys are dropped from the incoming dictionary to the outgoing
    one.
    """
    extra_vars = get_install_extra_vars(json)
    # The extra_vars dictionary gets updated here *with whatever exists in the
    # json* so any modifications must happen after this is performed to ensure
    # that changes are not overwritten.
    extra_vars.update(json)

    # enforce this option, which is the only "scenario" supported
    extra_vars['raw_multi_journal'] = True

    device_map = json["devices"]
    devices = []
    journal_devices = []

    for device, journal in device_map.items():
        devices.append(device)
        journal_devices.append(journal)

    # add the corresponding device and journal to what ceph-ansible expects
    extra_vars['devices'] = devices
    extra_vars['raw_journal_devices'] = journal_devices

    monitor_hosts = [mon['host'] for mon in json["monitors"]]
    host = json['host']
    if host in monitor_hosts:
        logger.info("The host %s is in both the OSD and MON groups.", host)
        # The role ceph-mon isn't used during on OSD configure so the
        # monitor_name variable is never set, but it needs to exist because
        # the ceph-common role will attempt to restart mons because this host
        # is both a mon and an osd.
        extra_vars['monitor_name'] = host

    if "cluster_name" in extra_vars:
        extra_vars["cluster"] = extra_vars["cluster_name"]
        del extra_vars["cluster_name"]

    # These are items that came via the JSON that should never be passed into
    # ceph-ansible because they are no longer needed
    del extra_vars['host']
    del extra_vars['monitors']
    if 'redhat_storage' in json:
        del extra_vars['redhat_storage']

    return extra_vars



# -*- coding: utf-8 -*-
"""
    pygments.styles.solarized.light
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    The Solarized style, inspired by Schoonover.

    :copyright: Copyright 2012 by the Shoji KUMAGAI, see AUTHORS.
    :license: MIT, see LICENSE for details.
"""

from pygments.style import Style
from pygments.token import Keyword, Name, Comment, String, Error, Text, \
     Number, Operator, Generic, Whitespace, Other, Literal, Punctuation


class LightStyle(Style):
    """
    The Solarized Light style, inspired by Schoonover.
    """
    background_color = '#fdf6e3'
    default_style = ""

    styles = {
        Text:                   '#657b83',             # base00 ; class: ''
        Whitespace:             '#fdf6e3',             # base3  ; class: 'w'
        Error:                  '#dc322f',             # red    ; class: 'err'
        Other:                  '#657b83',             # base00 ; class: 'x'

        Comment:                'italic #93a1a1',      # base1  ; class: 'c'
        Comment.Multiline:      'italic #93a1a1',      # base1  ; class: 'cm'
        Comment.Preproc:        'italic #93a1a1',      # base1  ; class: 'cp'
        Comment.Single:         'italic #93a1a1',      # base1  ; class: 'c1'
        Comment.Special:        'italic #93a1a1',      # base1  ; class: 'cs'

        Keyword:                '#859900',             # green  ; class: 'k'
        Keyword.Constant:       '#859900',             # green  ; class: 'kc'
        Keyword.Declaration:    '#859900',             # green  ; class: 'kd'
        Keyword.Namespace:      '#cb4b16',             # orange ; class: 'kn'
        Keyword.Pseudo:         '#cb4b16',             # orange ; class: 'kp'
        Keyword.Reserved:       '#859900',             # green  ; class: 'kr'
        Keyword.Type:           '#859900',             # green  ; class: 'kt'

        Operator:               '#657b83',             # base00 ; class: 'o'
        Operator.Word:          '#859900',             # green  ; class: 'ow'

        Name:                   '#586e75',             # base01 ; class: 'n'
        Name.Attribute:         '#657b83',             # base00 ; class: 'na'
        Name.Builtin:           '#268bd2',             # blue   ; class: 'nb'
        Name.Builtin.Pseudo:    'bold #268bd2',        # blue   ; class: 'bp'
        Name.Class:             '#268bd2',             # blue   ; class: 'nc'
        Name.Constant:          '#b58900',             # yellow ; class: 'no'
        Name.Decorator:         '#cb4b16',             # orange ; class: 'nd'
        Name.Entity:            '#cb4b16',             # orange ; class: 'ni'
        Name.Exception:         '#cb4b16',             # orange ; class: 'ne'
        Name.Function:          '#268bd2',             # blue   ; class: 'nf'
        Name.Property:          '#268bd2',             # blue   ; class: 'py'
        Name.Label:             '#657b83',             # base00 ; class: 'nc'
        Name.Namespace:         '#b58900',             # yellow ; class: 'nn'
        Name.Other:             '#657b83',             # base00 ; class: 'nx'
        Name.Tag:               '#859900',             # green  ; class: 'nt'
        Name.Variable:          '#cb4b16',             # orange ; class: 'nv'
        Name.Variable.Class:    '#268bd2',             # blue   ; class: 'vc'
        Name.Variable.Global:   '#268bd2',             # blue   ; class: 'vg'
        Name.Variable.Instance: '#268bd2',             # blue   ; class: 'vi'

        Number:                 '#2aa198',             # cyan   ; class: 'm'
        Number.Float:           '#2aa198',             # cyan   ; class: 'mf'
        Number.Hex:             '#2aa198',             # cyan   ; class: 'mh'
        Number.Integer:         '#2aa198',             # cyan   ; class: 'mi'
        Number.Integer.Long:    '#2aa198',             # cyan   ; class: 'il'
        Number.Oct:             '#2aa198',             # cyan   ; class: 'mo'

        Literal:                '#657b83',             # base00 ; class: 'l'
        Literal.Date:           '#657b83',             # base00 ; class: 'ld'

        Punctuation:            '#657b83',             # base00 ; class: 'p'

        String:                 '#2aa198',             # cyan   ; class: 's'
        String.Backtick:        '#2aa198',             # cyan   ; class: 'sb'
        String.Char:            '#2aa198',             # cyan   ; class: 'sc'
        String.Doc:             '#2aa198',             # cyan   ; class: 'sd'
        String.Double:          '#2aa198',             # cyan   ; class: 's2'
        String.Escape:          '#cb4b16',             # orange ; class: 'se'
        String.Heredoc:         '#2aa198',             # cyan   ; class: 'sh'
        String.Interpol:        '#cb4b16',             # orange ; class: 'si'
        String.Other:           '#2aa198',             # cyan   ; class: 'sx'
        String.Regex:           '#2aa198',             # cyan   ; class: 'sr'
        String.Single:          '#2aa198',             # cyan   ; class: 's1'
        String.Symbol:          '#2aa198',             # cyan   ; class: 'ss'

        Generic:                '#657b83',             # base00 ; class: 'g'
        Generic.Deleted:        '#657b83',             # base00 ; class: 'gd'
        Generic.Emph:           '#657b83',             # base00 ; class: 'ge'
        Generic.Error:          '#657b83',             # base00 ; class: 'gr'
        Generic.Heading:        '#657b83',             # base00 ; class: 'gh'
        Generic.Inserted:       '#657b83',             # base00 ; class: 'gi'
        Generic.Output:         '#657b83',             # base00 ; class: 'go'
        Generic.Prompt:         '#657b83',             # base00 ; class: 'gp'
        Generic.Strong:         '#657b83',             # base00 ; class: 'gs'
        Generic.Subheading:     '#657b83',             # base00 ; class: 'gu'
        Generic.Traceback:      '#657b83',             # base00 ; class: 'gt'
    }



import logging

from datetime import datetime
from celery import shared_task
from ceph_installer import util, models, process


logger = logging.getLogger(__name__)


@shared_task
def call_ansible(inventory, identifier, tags="", skip_tags="", extra_vars=None, playbook="site.yml.sample", **kw):
    """
    This task builds an ansible-playbook command and runs it.

    ``inventory``: A list of tuples that details an ansible inventory. For example:
                   [('mons', ['mon1.host', 'mon2.host']), ('osds', ['osd1.host'])]
    ``tags``: The tags as a comma-delimeted string that represents all the tags
              this ansible call should follow. For example "package-install, other-tag"
    ``skip_tags``: The tags as a comma-delimeted string that represents all the tags
                   this ansible call should skip. For example "package-install, other-tag"
    ``identifier``: The UUID identifer for the task object so this function can capture process
                    metadata and persist it to the database
    ``verbose``: Optional keyword argument, to flag the need for increased verbosity
                 when running ansible
    """
    verbose = kw.get('verbose', False)
    if not extra_vars:
        extra_vars = dict()
    hosts_file = util.generate_inventory_file(inventory, identifier)
    command = process.make_ansible_command(
        hosts_file, identifier, tags=tags, extra_vars=extra_vars,
        skip_tags=skip_tags, playbook=playbook, verbose=verbose
    )
    task = models.Task.query.filter_by(identifier=identifier).first()
    task.command = ' '.join(command)
    task.started = datetime.now()
    # force a commit here so we can reference this command later if it fails
    models.commit()
    working_dir = util.get_ceph_ansible_path()
    # ansible depends on relative pathing to figure out how to load
    # plugins, among other things. Setting the current working directory
    # for this subprocess call to the directory where the playbook resides
    # allows ansible to properly find action plugins defined in ceph-ansible.
    kwargs = dict(cwd=working_dir)
    try:
        out, err, exit_code = process.run(command, **kwargs)
    except Exception as error:
        task.succeeded = False
        task.exit_code = -1
        task.stderr = str(error)
        logger.exception('failed to run command')
    else:
        task.succeeded = not exit_code
        task.exit_code = exit_code
        task.stdout = out
        task.stderr = err

    task.ended = datetime.now()
    models.commit()



from ceph_installer.controllers import agent


class TestAgentController(object):

    def test_index_get_is_not_allowed(self, session):
        result = session.app.get("/api/agent/", expect_errors=True)
        assert result.status_int == 405

    def test_index_incorrect_schema(self, session):
        data = dict(hosts=["google.com"], bogus="foo")
        result = session.app.post_json("/api/agent/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_index_post_works_with_right_schema(self, session, monkeypatch):
        monkeypatch.setattr(agent.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = dict(hosts=["node1"])
        result = session.app.post_json("/api/agent/", params=data)
        assert result.json['endpoint'] == '/api/agent/'
        assert result.json['identifier'] is not None

    def test_invalid_master_value_is_detected(self, session, monkeypatch):
        data = dict(hosts=["google.com"], master=1)
        result = session.app.post_json("/api/agent/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert result.json['message'].endswith('not of type string')

    def test_master_is_defined_and_used(self, session, monkeypatch, argtest):
        monkeypatch.setattr(agent.call_ansible, 'apply_async', argtest)
        data = dict(hosts=["node1"], master='installer.host')
        session.app.post_json("/api/agent/", params=data)
        # argtest.kwargs['kwargs'] may look a bit redundant, but celery tasks
        # require that same signature as the 'recorder' fixture we have to
        # capture args and kwargs
        ansible_extra_args = argtest.kwargs['kwargs']['extra_vars']
        assert ansible_extra_args['agent_master_host'] == 'installer.host'

    def test_master_is_not_defined_and_is_inferred(self, session, monkeypatch, argtest):
        monkeypatch.setattr(agent.call_ansible, 'apply_async', argtest)
        data = dict(hosts=["node1"])
        session.app.post_json("/api/agent/", params=data)
        ansible_extra_args = argtest.kwargs['kwargs']['extra_vars']
        assert ansible_extra_args['agent_master_host'] == 'localhost'


class TestAgentVerbose(object):

    def test_install_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            agent.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"], "verbose": True}
        session.app.post_json("/api/agent/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is True

    def test_install_non_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            agent.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"]}
        session.app.post_json("/api/agent/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is False



from sqlalchemy import Column, Integer, String,  Boolean, DateTime, UnicodeText
from sqlalchemy.orm.exc import DetachedInstanceError
from ceph_installer.models import Base


class Task(Base):

    __tablename__ = 'tasks'
    id = Column(Integer, primary_key=True)
    identifier = Column(String(256), unique=True, nullable=False, index=True)
    endpoint = Column(String(256), index=True)
    user_agent = Column(String(512))
    request = Column(UnicodeText)
    http_method = Column(String(64))
    command = Column(String(256))
    stderr = Column(UnicodeText)
    stdout = Column(UnicodeText)
    started = Column(DateTime)
    ended = Column(DateTime)
    succeeded = Column(Boolean(), default=False)
    exit_code = Column(Integer)

    def __init__(self, request=None, **kw):
        self._extract_request_metadata(request)
        for k, v in kw.items():
            setattr(self, k, v)

    def _extract_request_metadata(self, request):
        self.http_method = getattr(request, 'method', '')
        self.request = str(getattr(request, 'body', ''))
        self.user_agent = getattr(request, 'user_agent', '')

    def __repr__(self):
        try:
            return '<Task %r>' % self.identifier
        except DetachedInstanceError:
            return '<Task detached>'

    def __json__(self):
        return dict(
            identifier = self.identifier,
            endpoint = self.endpoint,
            command = self.command,
            stderr = self.stderr,
            stdout = self.stdout,
            started = self.started,
            ended = self.ended,
            succeeded = self.succeeded,
            exit_code = self.exit_code,
            user_agent = self.user_agent,
            request = self.request,
            http_method = self.http_method,
        )



import pytest
from ceph_installer import hooks
from sqlalchemy.exc import OperationalError


class FakeState(object):

    def __init__(self, **kw):
        for k, v in kw.items():
            setattr(self, k, v)


class TestAnsibleExists(object):

    def test_ansible_is_there(self, monkeypatch):
        monkeypatch.setattr(
            hooks, 'which',
            lambda x: '/usr/bin/ansible-playbook')
        assert hooks.ansible_exists() is None

    def test_ansible_is_not_there(self, monkeypatch):
        monkeypatch.setattr(
            hooks, 'which',
            lambda x: None)
        with pytest.raises(hooks.SystemCheckError):
            hooks.ansible_exists()


class TestCeleryHasWorkers(object):

    def test_celery_has_workers(self, monkeypatch):
        stats = lambda: {'value': 'some stat'}
        monkeypatch.setattr(
            hooks, 'inspect',
            lambda: FakeState(stats=stats))
        assert hooks.celery_has_workers() is None

    def test_celery_has_no_workers(self, monkeypatch):
        monkeypatch.setattr(
            hooks, 'inspect',
            lambda: FakeState(stats=lambda: None))
        with pytest.raises(hooks.SystemCheckError):
            hooks.celery_has_workers()


class TestRabbitMQIsRunning(object):

    def test_is_running(self, monkeypatch):
        def errors(): raise IOError
        monkeypatch.setattr(
            hooks, 'celery_has_workers',
            errors)
        with pytest.raises(hooks.SystemCheckError):
            assert hooks.rabbitmq_is_running()

    def test_is_not_running(self, monkeypatch):
        monkeypatch.setattr(
            hooks, 'celery_has_workers',
            lambda: None)
        assert hooks.rabbitmq_is_running() is None


class TestDBConnection(object):

    def test_is_connected(self, monkeypatch):
        monkeypatch.setattr(
            hooks.models, 'Task',
            FakeState(get=lambda x: None))
        assert hooks.database_connection() is None

    def test_is_not_running(self, monkeypatch):
        def errors(x): raise OperationalError(None, None, None, None)
        monkeypatch.setattr(
            hooks.models, 'Task',
            FakeState(get=errors))
        with pytest.raises(hooks.SystemCheckError):
            hooks.database_connection() is None



from pecan import expose

from ceph_installer.models import Task
from ceph_installer.controllers import error


class TaskController(object):

    def __init__(self, task_id):
        self.task = Task.query.filter_by(identifier=task_id).first()
        if not self.task:
            error(404, '%s is not available' % task_id)

    @expose('json')
    def index(self):
        return self.task


class TasksController(object):

    @expose('json')
    def index(self):
        return Task.query.all()

    @expose('json')
    def install(self):
        return {}

    @expose('json')
    def configure(self):
        return {}

    @expose()
    def _lookup(self, task_id, *remainder):
        return TaskController(task_id), remainder



from notario.validators import types, recursive
from notario.utils import forced_leaf_validator
from notario.exceptions import Invalid
from notario.decorators import optional


def list_of_hosts(value):
    assert isinstance(value, list), "requires format: ['host1', 'host2']"


def list_of_devices(value):
    assert isinstance(value, list), "requires format: ['/dev/sdb', '/dev/sdc']"


@forced_leaf_validator
def devices_object(_object, *args):
    error_msg = 'not of type dictionary'
    try:
        assert isinstance(_object, dict)
    except AssertionError:
        if args:
            raise Invalid('dict type', pair='value', msg=None, reason=error_msg, *args)
        raise

    v = recursive.AllObjects((types.string, types.string))
    # this is truly unfortunate but we don't have access to the 'tree' here
    # (the tree is the path to get to the failing key. We settle by just being
    # able to report nicely.
    v(_object, [])


def list_of_monitors(value):
    msg = 'requires format: [{"host": "mon1.host", "interface": "eth1"},{"host": "mon2.host", "address": "10.0.0.1"}]'
    assert isinstance(value, list), msg
    msg = 'address or interface is required for monitor lists: [{"host": "mon1", "interface": "eth1", {"host": "mon2", "address": "10.0.0.1"}]'
    for monitor in value:
        assert isinstance(monitor, dict), msg
        assert "host" in monitor, msg
        try:
            assert "interface" in monitor, msg
        except AssertionError:
            assert "address" in monitor, msg


conf = (
    (optional("global"), types.dictionary),
    (optional("mds"), types.dictionary),
    (optional("mon"), types.dictionary),
    (optional("osd"), types.dictionary),
    (optional("rgw"), types.dictionary),
)

install_schema = (
    ("hosts", list_of_hosts),
    (optional("redhat_storage"), types.boolean),
    (optional("redhat_use_cdn"), types.boolean),
    (optional("verbose"), types.boolean),
)

agent_install_schema = (
    ("hosts", list_of_hosts),
    (optional("master"), types.string),
    (optional("verbose"), types.boolean),
)

mon_install_schema = (
    (optional("calamari"), types.boolean),
    ("hosts", list_of_hosts),
    (optional("redhat_storage"), types.boolean),
    (optional("redhat_use_cdn"), types.boolean),
    (optional("verbose"), types.boolean),
)

mon_configure_schema = (
    (optional("address"), types.string),
    (optional("calamari"), types.boolean),
    (optional("cluster_name"), types.string),
    (optional("cluster_network"), types.string),
    (optional("conf"), conf),
    ("fsid", types.string),
    ("host", types.string),
    (optional("interface"), types.string),
    ("monitor_secret", types.string),
    (optional("monitors"), list_of_monitors),
    ("public_network", types.string),
    (optional("redhat_storage"), types.boolean),
    (optional("verbose"), types.boolean),
)

osd_configure_schema = (
    (optional("cluster_name"), types.string),
    (optional("cluster_network"), types.string),
    (optional("conf"), conf),
    ("devices", devices_object),
    ("fsid", types.string),
    ("host", types.string),
    ("journal_size", types.integer),
    ("monitors", list_of_monitors),
    ("public_network", types.string),
    (optional("redhat_storage"), types.boolean),
    (optional("verbose"), types.boolean),
)



from pecan import response, expose
from ceph_installer.hooks import system_checks, SystemCheckError


class StatusController(object):

    @expose('json')
    def index(self):
        for check in system_checks:
            try:
                check()
            except SystemCheckError as system_error:
                response.status = 500
                return {'message': system_error.message}
        return dict(message="ok")



import os
import tempfile
from pecan.testing import load_test_app

from copy import deepcopy
from pecan import conf
from pecan import configuration
from sqlalchemy import create_engine
from sqlalchemy.pool import NullPool

from ceph_installer import models as _db
import pytest

os.environ['HOME'] = tempfile.mkdtemp(suffix='.ceph-installer-home')

DBNAME = 'ceph_installertest.db'
BIND = 'sqlite:///' + os.environ['HOME']


def config_file():
    here = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(here, 'config.py')


@pytest.fixture
def fake():
    class Fake(object):
        def __init__(self, *a, **kw):
            for k, v, in kw.items():
                setattr(self, k, v)
    return Fake


@pytest.fixture
def argtest():
    """
    Simple helper to use with monkeypatch so that a callable can be inspected
    afterwards to see if it was called with certain arguments.
    """
    class TestArgs(object):
        def __call__(self, *args, **kwargs):
            self.args = list(args)
            self.kwargs = kwargs
    return TestArgs()


@pytest.fixture(scope='session')
def app(request):
    config = configuration.conf_from_file(config_file()).to_dict()

    # Add the appropriate connection string to the app config.
    config['sqlalchemy'] = {
        'url': '%s/%s' % (BIND, DBNAME),
        'echo':          True,
        'echo_pool':     True,
        'pool_recycle':  3600,
        'encoding':      'utf-8'
    }

    # Set up a fake app
    app = TestApp(load_test_app(config))
    return app


@pytest.fixture(scope='session')
def connection(app, request):
    """Session-wide test database."""
    # Connect and create the temporary database
    print "=" * 80
    print "CREATING TEMPORARY DATABASE FOR TESTS"
    print "=" * 80

    # Bind and create the database tables
    _db.clear()
    engine_url = '%s/%s' % (BIND, DBNAME)

    db_engine = create_engine(
        engine_url,
        encoding='utf-8',
        poolclass=NullPool)

    # AKA models.start()
    _db.Session.bind = db_engine
    _db.metadata.bind = _db.Session.bind

    _db.Base.metadata.create_all(db_engine)
    _db.commit()
    _db.clear()

    #connection = db_engine.connect()

    def teardown():
        _db.Base.metadata.drop_all(db_engine)

    request.addfinalizer(teardown)

    # Slap our test app on it
    _db.app = app
    return _db


@pytest.fixture(scope='function')
def session(connection, request):
    """Creates a new database session for a test."""
    _config = configuration.conf_from_file(config_file()).to_dict()
    config = deepcopy(_config)

    # Add the appropriate connection string to the app config.
    config['sqlalchemy'] = {
        'url': '%s/%s' % (BIND, DBNAME),
        'encoding': 'utf-8',
        'poolclass': NullPool
    }

    connection.start()

    def teardown():
        from sqlalchemy.engine import reflection

        # Tear down and dispose the DB binding
        connection.clear()

        # start a transaction
        engine = conf.sqlalchemy.engine
        conn = engine.connect()
        trans = conn.begin()

        inspector = reflection.Inspector.from_engine(engine)

        # gather all data first before dropping anything.
        # some DBs lock after things have been dropped in
        # a transaction.
        conn.execute("DELETE FROM  %s;" % (
            ', '.join(inspector.get_table_names())
        ))

        trans.commit()
        conn.close()

    request.addfinalizer(teardown)
    return connection


class TestApp(object):
    """
    A controller test starts a database transaction and creates a fake
    WSGI app.
    """

    __headers__ = {}

    def __init__(self, app):
        self.app = app

    def _do_request(self, url, method='GET', **kwargs):
        methods = {
            'GET': self.app.get,
            'POST': self.app.post,
            'POSTJ': self.app.post_json,
            'PUT': self.app.put,
            'DELETE': self.app.delete
        }
        kwargs.setdefault('headers', {}).update(self.__headers__)
        return methods.get(method, self.app.get)(str(url), **kwargs)

    def post_json(self, url, **kwargs):
        """
        @param (string) url - The URL to emulate a POST request to
        @returns (paste.fixture.TestResponse)
        """
        return self._do_request(url, 'POSTJ', **kwargs)

    def post(self, url, **kwargs):
        """
        @param (string) url - The URL to emulate a POST request to
        @returns (paste.fixture.TestResponse)
        """
        return self._do_request(url, 'POST', **kwargs)

    def get(self, url, **kwargs):
        """
        @param (string) url - The URL to emulate a GET request to
        @returns (paste.fixture.TestResponse)
        """
        return self._do_request(url, 'GET', **kwargs)

    def put(self, url, **kwargs):
        """
        @param (string) url - The URL to emulate a PUT request to
        @returns (paste.fixture.TestResponse)
        """
        return self._do_request(url, 'PUT', **kwargs)

    def delete(self, url, **kwargs):
        """
        @param (string) url - The URL to emulate a DELETE request to
        @returns (paste.fixture.TestResponse)
        """
        return self._do_request(url, 'DELETE', **kwargs)


# this console logging configuration is basically just to be able to see output
# in tests, and this file gets executed by py.test when it runs, so we get that
# for free.
import logging
# Console Logger
sh = logging.StreamHandler()
sh.setLevel(logging.WARNING)

formatter = logging.Formatter(
    fmt='%(asctime)s.%(msecs)03d %(process)d:%(levelname)s:%(name)s:%(message)s',
    datefmt='%Y-%m-%dT%H:%M:%S',
)
sh.setFormatter(formatter)


# because we're in a module already, __name__ is not the ancestor of
# the rest of the package; use the root as the logger for everyone
root_logger = logging.getLogger()

# allow all levels at root_logger, handlers control individual levels
root_logger.setLevel(logging.DEBUG)
root_logger.addHandler(sh)

console_loglevel = logging.DEBUG  # start at DEBUG for now

# Console Logger
sh.setLevel(console_loglevel)



import os
import subprocess
import tempfile
import logging
import json

from ceph_installer.util import which, get_ceph_ansible_path

logger = logging.getLogger(__name__)


def make_ansible_command(hosts_file, identifier, extra_vars=None, tags='', skip_tags='', playbook='site.yml.sample', **kw):
    """
    This utility will compose the command needed to run ansible, capture its
    stdout and stderr to a file
    ``verbose``: Optional keyword argument, to flag the need for increased
                 verbosity when running ansible
    """
    if kw.get('verbose'):
        verbose_flag = '-vvvv'
    else:
        verbose_flag = '-v'
    ceph_ansible_path = get_ceph_ansible_path()
    playbook = os.path.join(ceph_ansible_path, playbook)
    ansible_path = which('ansible-playbook')
    if not extra_vars:
        extra_vars = dict()

    extra_vars = json.dumps(extra_vars)

    cmd = [
        ansible_path, verbose_flag, '-u', 'ceph-installer',
        playbook, '-i', hosts_file, "--extra-vars", extra_vars,
    ]

    if tags:
        cmd.extend(['--tags', tags])

    if skip_tags:
        cmd.extend(['--skip-tags', skip_tags])

    return cmd


def temp_file(identifier, std):
    return tempfile.NamedTemporaryFile(
        prefix="{identifier}_{std}".format(
            identifier=identifier, std=std), delete=False
    )


def run(arguments, send_input=None, **kwargs):
    """
    A small helper to run a system command using ``subprocess.Popen``.

    This returns the output of the command and the return code of the process
    in a tuple::

        (stdout, stderr, returncode)
    """
    env = os.environ.copy()
    env["ANSIBLE_HOST_KEY_CHECKING"] = "False"
    logger.info('Running command: %s' % ' '.join(arguments))
    process = subprocess.Popen(
        arguments,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        env=env,
        **kwargs)
    if send_input:
        out, err = process.communicate(send_input)
    else:
        out, err = process.communicate()

    # Ensure we are dealing with strings (might get a None) and decode them
    out = str(out).decode('utf-8', 'replace')
    err = str(err).decode('utf-8', 'replace')
    return out, err, process.returncode



from ceph_installer import hooks

# Server Specific Configurations
server = {
    'port': '8181',
    'host': '0.0.0.0'
}

# Pecan Application Configurations
app = {
    'root': 'ceph_installer.controllers.root.RootController',
    'modules': ['ceph_installer'],
    'debug': False,
    'hooks': [hooks.CustomErrorHook()]
}

logging = {
    'root': {'level': 'INFO', 'handlers': ['console']},
    'loggers': {
        'ceph_installer': {'level': 'DEBUG', 'handlers': ['console'], 'propagate': False},
        'pecan': {'level': 'INFO', 'handlers': ['console'], 'propagate': False},
        'py.warnings': {'handlers': ['console']},
        '__force_dict__': True
    },
    # XXX Determine the right location for logs
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        }
    },
    'formatters': {
        'simple': {
            'format': ('%(asctime)s %(levelname)-5.5s [%(name)s]'
                       '[%(threadName)s] %(message)s')
        },
        'color': {
            '()': 'pecan.log.ColorFormatter',
            'format': ('%(asctime)s [%(padded_color_levelname)s] [%(name)s]'
                       '[%(threadName)s] %(message)s'),
        '__force_dict__': True
        }
    }
}


sqlalchemy = {
    # XXX Determine the right location for the database
    'url': 'sqlite:////var/lib/ceph-installer/ceph_installer.db',
    'echo':          True,
    'echo_pool':     True,
    'pool_recycle':  3600,
    'encoding':      'utf-8'
}



from os import path
from textwrap import dedent
from tambo import Transport

from ceph_installer import process
from ceph_installer.cli import log

this_dir = path.abspath(path.dirname(__file__))
top_dir = path.dirname(path.dirname(this_dir))
playbook_path = path.join(top_dir, 'deploy/playbooks')


class Dev(object):

    help = "Development options"
    options = ['--user', '--branch']
    _help = dedent("""
    Deploying the ceph-installer HTTP service to a remote server with ansible.
    This command wraps ansible and certain flags to make it easier to deploy
    a development version.

    Usage::

        ceph-installer dev $HOST

    Note: Requires a remote user with passwordless sudo. User defaults to
    "vagrant".

    Options:

    --user        Define a user to connect to the remote server. Defaults  to 'vagrant'
    --branch      What branch to use for the deployment. Defaults to 'master'
    -vvvv         Enable high verbosity when running ansible
    """)

    def __init__(self, arguments):
        self.arguments = arguments

    def main(self):
        parser = Transport(self.arguments, options=self.options, check_help=True)
        parser.catch_help = self._help
        parser.parse_args()
        parser.catches_help()
        branch = parser.get('--branch', 'master')
        user = parser.get('--user', 'vagrant')
        high_verbosity = '-vvvv' if parser.has('-vvvv') else '-v'
        if not parser.unknown_commands:
            log.error("it is required to pass a host to deploy to, but none was provided")
            raise SystemExit(1)

        command = [
            "ansible-playbook",
            "-i", "%s," % parser.unknown_commands[-1],
            high_verbosity,
            "-u", user,
            "--extra-vars", 'branch=%s' % branch,
            "deploy.yml",
        ]
        log.debug("Running command: %s" % ' '.join(command))
        out, err, code = process.run(command, cwd=playbook_path)
        log.error(err)
        log.debug(out)



import sys

from tambo import Transport
import ceph_installer
from ceph_installer.cli import log
from ceph_installer.cli import dev, task, constants
from ceph_installer.cli.decorators import catches


class CephInstaller(object):
    _help = """
A command line utility to install and configure Ceph using an HTTP API as a REST service
to call Ansible.

Address: %s
Version: %s

Global Options:
-h, --help, help    Show this program's help menu
--log, --logging    Set the level of logging. Acceptable values:
                    debug, warning, error, critical

Environment Variables:
CEPH_INSTALLER_ADDRESS    Define the location of the installer.
                          Defaults to "http://localhost:8181"

%s
    """

    mapper = {'dev': dev.Dev, 'task': task.Task}

    def __init__(self, argv=None, parse=True):
        self.plugin_help = "No plugins found/loaded"
        if argv is None:
            argv = sys.argv
        if parse:
            self.main(argv)

    def help(self, subhelp):
        version = ceph_installer.__version__
        return self._help % (
            constants.server_address, version,
            subhelp
        )

    @catches(KeyboardInterrupt, handle_all=True, logger=log)
    def main(self, argv):
        parser = Transport(argv, mapper=self.mapper,
                           options=[], check_help=False,
                           check_version=False)
        parser.parse_args()
        parser.catch_help = self.help(parser.subhelp())
        parser.catch_version = ceph_installer.__version__
        parser.mapper = self.mapper
        if len(argv) <= 1:
            return parser.print_help()
        parser.dispatch()
        parser.catches_help()
        parser.catches_version()



import os
from ceph_installer.controllers import setup


class TestSetupController(object):

    def test_index_generates_a_script(self, session):
        result = session.app.get('/setup/')
        assert '#!/bin/bash' in result.body

    def test_index_adds_the_right_endpoint_to_the_script(self, session):
        result = session.app.get('/setup/')
        assert 'http://localhost/setup/key/' in result.body

    def test_missing_ssh_directory(self, session, tmpdir, monkeypatch):
        rsa_path = os.path.join(str(tmpdir), '.ssh/id_rsa')
        monkeypatch.setattr(setup.os.path, 'expanduser', lambda x: rsa_path)
        result = session.app.get('/setup/key/', expect_errors=True)
        assert result.status_int == 500
        assert result.json['message'].startswith('.ssh directory not found')

    def test_missing_ssh_key(self, session, tmpdir, monkeypatch):
        tmpdir.mkdir('.ssh')
        rsa_path = os.path.join(str(tmpdir), '.ssh/id_rsa')
        monkeypatch.setattr(setup.os.path, 'expanduser', lambda x: rsa_path)
        result = session.app.get('/setup/key/', expect_errors=True)
        assert result.status_int == 500
        assert result.json['message'].startswith('expected public key not found')



import datetime
from sqlalchemy import create_engine, MetaData, event
from sqlalchemy.orm import scoped_session, sessionmaker, object_session, mapper
from sqlalchemy.ext.declarative import declarative_base
from pecan import conf


class _EntityBase(object):
    """
    A custom declarative base that provides some Elixir-inspired shortcuts.
    """

    @classmethod
    def filter_by(cls, *args, **kwargs):
        return cls.query.filter_by(*args, **kwargs)

    @classmethod
    def get(cls, *args, **kwargs):
        return cls.query.get(*args, **kwargs)

    def flush(self, *args, **kwargs):
        object_session(self).flush([self], *args, **kwargs)

    def delete(self, *args, **kwargs):
        object_session(self).delete(self, *args, **kwargs)

    def as_dict(self):
        return dict((k, v) for k, v in self.__dict__.items()
                    if not k.startswith('_'))

    def update_from_json(self, data):
        """
        We received a JSON blob with updated metadata information
        that needs to update some fields
        """
        for key in data.keys():
            setattr(self, key, data[key])


Session = scoped_session(sessionmaker())
metadata = MetaData()
Base = declarative_base(cls=_EntityBase)
Base.query = Session.query_property()


# Listeners:

@event.listens_for(mapper, 'init')
def auto_add(target, args, kwargs):
    Session.add(target)


def update_timestamp(mapper, connection, target):
    """
    Automate the 'modified' attribute when a model changes
    """
    target.modified = datetime.datetime.utcnow()


# Utilities:

def get_or_create(model, **kwargs):
    instance = model.filter_by(**kwargs).first()
    if instance:
        return instance
    else:
        instance = model(**kwargs)
        commit()
        return instance


def init_model():
    """
    This is a stub method which is called at application startup time.

    If you need to bind to a parse database configuration, set up tables or
    ORM classes, or perform any database initialization, this is the
    recommended place to do it.

    For more information working with databases, and some common recipes,
    see http://pecan.readthedocs.org/en/latest/databases.html

    For creating all metadata you would use::

        Base.metadata.create_all(conf.sqlalchemy.engine)

    """
    conf.sqlalchemy.engine = _engine_from_config(conf.sqlalchemy)
    Session.configure(bind=conf.sqlalchemy.engine)


def _engine_from_config(configuration):
    configuration = dict(configuration)
    url = configuration.pop('url')
    return create_engine(url, **configuration)


def start():
    Session()
    metadata.bind = conf.sqlalchemy.engine


def start_read_only():
    start()


def commit():
    Session.commit()


def rollback():
    Session.rollback()


def clear():
    Session.remove()
    Session.close()


def flush():
    Session.flush()


from tasks import Task  # noqa



import logging

from pecan import expose, request
from pecan.ext.notario import validate
from uuid import uuid4

from ceph_installer.controllers import error
from ceph_installer.tasks import call_ansible
from ceph_installer import schemas
from ceph_installer import models
from ceph_installer import util

logger = logging.getLogger(__name__)


class OSDController(object):

    @expose('json')
    def index(self):
        # TODO: allow some autodiscovery here so that clients can see what is
        # available
        return dict()

    @expose(generic=True, template='json')
    def install(self):
        error(405)

    @install.when(method='POST', template='json')
    @validate(schemas.install_schema, handler="/errors/schema")
    def install_post(self):
        hosts = request.json.get('hosts')
        verbose_ansible = request.json.get('verbose', False)
        extra_vars = util.get_install_extra_vars(request.json)
        identifier = str(uuid4())
        task = models.Task(
            request=request,
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(
            extra_vars=extra_vars,
            tags="package-install",
            verbose=verbose_ansible,
        )
        call_ansible.apply_async(
            args=([('osds', hosts)], identifier),
            kwargs=kwargs
        )

        return task

    @expose(generic=True, template='json')
    def configure(self):
        error(405)

    @configure.when(method='POST', template='json')
    @validate(schemas.osd_configure_schema, handler="/errors/schema")
    def configure_post(self):
        hosts = [request.json['host']]
        monitor_hosts = util.parse_monitors(request.json["monitors"])
        verbose_ansible = request.json.get('verbose', False)
        # even with configuring we need to tell ceph-ansible
        # if we're working with upstream ceph or red hat ceph storage
        extra_vars = util.get_osd_configure_extra_vars(request.json)
        if 'verbose' in extra_vars:
            del extra_vars['verbose']
        if 'conf' in extra_vars:
            extra_vars['ceph_conf_overrides'] = request.json['conf']
            del extra_vars['conf']
        identifier = str(uuid4())
        task = models.Task(
            request=request,
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(
            extra_vars=extra_vars,
            skip_tags="package-install",
            playbook="osd-configure.yml",
            verbose=verbose_ansible,
        )
        call_ansible.apply_async(
            args=([('osds', hosts), ('mons', monitor_hosts)], identifier),
            kwargs=kwargs
        )

        return task



"""
This is really not a logging facility. For the CLI we only require a nice color
output on some interactions. Full logging is not needed.
"""
import sys
from ceph_installer.cli import util


def error(message):
    line = "%s %s\n" % (util.red_arrow, message)
    sys.stderr.write(line)


def debug(message):
    line = "%s %s\n" % (util.blue_arrow, message)
    sys.stdout.write(line)


def info(message):
    line = "%s %s\n" % (util.bold_arrow, message)
    sys.stdout.write(line)


def warning(message):
    line = "%s %s\n" % (util.yellow_arrow, message)
    sys.stderr.write(line)



from os import path
import sys
import requests
import time
from textwrap import dedent
from tambo import Transport

from ceph_installer.cli import log, constants


class Task(object):

    help = "/api/tasks/ operations"
    options = []
    _help = dedent("""
    Human-readable task information: stdout, stderr, and the ability to "poll"
    a task that waits until the command completes to be able to show the output
    in a readable way.

    Usage::

        ceph-installer task $IDENTIFIER

    Options:

    --poll        Poll until the task has completed (either on failure or success)
    stdout        Retrieve the stdout output from the task
    stderr        Retrieve the stderr output from the task
    command       The actual command used to call ansible
    ended         The timestamp (in UTC) when the command completed
    started       The timestamp (in UTC) when the command started
    exit_code     The shell exit status for the process
    succeeded     Boolean value to indicate if process completed correctly
    """)

    def __init__(self, arguments):
        self.arguments = arguments
        self.tasks_url = path.join(constants.server_address, 'api/tasks')
        self.identifier = ''

    @property
    def request_url(self):
        url = path.join(self.tasks_url, self.identifier)
        # and add a trailing slash so that the request is done at the correct
        # canonical url
        if not url.endswith('/'):
            url = "%s/" % url
        return url

    def get(self, key):
        """
        :arg key: any actual key that can be present in the JSON output
        """
        log.info('requesting task at: %s' % self.request_url)
        response = requests.get(self.request_url)
        json = response.json()
        if response.status_code >= 400:
            return log.error(json['message'])
        try:
            value = json[key]
            log.info("%s: %s" % (key, value))
        except KeyError:
            return log.warning('no value found for: "%s"' % key)

    def summary(self):
        response = requests.get(self.request_url)
        json = response.json()
        if response.status_code >= 400:
            log.error(json['message'])
        for k, v in json.items():
            log.debug("%s: %s" % (k, v))

    def process_response(self, silent=False):
        response = requests.get(self.request_url)
        json = response.json()
        if response.status_code >= 400:
            if not silent:
                log.error(json['message'])
            return {}
        return json

    def poll(self):
        log.info('Polling task at: %s' % self.request_url)
        json = self.process_response()
        # JSON could be set to None
        completed = json.get('ended', False)
        while not completed:
            sequence = ['.', '..', '...', '....']
            for s in sequence:
                sys.stdout.write('\r' + ' '*80)
                string = "Waiting for completed task%s" % s
                sys.stdout.write('\r' + string)
                time.sleep(0.3)
                sys.stdout.flush()
            json = self.process_response(silent=True)
            completed = json.get('ended', False)
        sys.stdout.write('\r' + ' '*80)
        sys.stdout.flush()
        sys.stdout.write('\r'+'Task Completed!\n')
        for k, v in json.items():
            log.debug("%s: %s" % (k, v))

    def main(self):
        parser = Transport(self.arguments, options=self.options, check_help=True)
        parser.catch_help = self._help
        parser.parse_args()
        parser.catches_help()
        if not parser.unknown_commands:
            log.error("it is required to pass an identifer, but none was provided")
            raise SystemExit(1)
        self.identifier = parser.unknown_commands[-1]
        if parser.has('--poll'):
            return self.poll()

        for key in [
                'stdout', 'stderr', 'command', 'ended',
                'started', 'succeeded', 'exit_code']:
            if parser.has(key):
                return self.get(key)

        # if nothing else matches, just try to give a generic, full summary
        self.summary()



import logging
from pecan import expose, response, request


logger = logging.getLogger(__name__)
schema_logger = logging.getLogger("%s.schema" % __name__)


class ErrorController(object):

    @expose('json')
    def schema(self, **kw):
        response.status = 400
        schema_logger.error(request.validation_error)
        try:
            path = request.validation_error._format_path()
            message = '%s%sfailed validation, %s' % (
                path,
                '' if path.endswith(' ') else ' ',
                request.validation_error.reason
            )
        except AttributeError:
            message = "invalid JSON was received"
        return dict(message=message)

    @expose('json')
    def invalid(self, **kw):
        msg = kw.get(
            'message',
            'invalid request'
        )
        response.status = 400
        return dict(message=msg)

    @expose('json')
    def forbidden(self, **kw):
        msg = kw.get(
            'message',
            'forbidden'
        )
        response.status = 403
        return dict(message=msg)

    @expose('json')
    def not_found(self, **kw):
        msg = kw.get(
            'message',
            'resource was not found'
        )
        response.status = 404
        return dict(message=msg)

    @expose('json')
    def not_allowed(self, **kw):
        msg = kw.get(
            'message',
            'method %s not allowed for "%s"' % (request.method, request.path)
        )
        response.status = 405
        return dict(message=msg)

    @expose('json')
    def unavailable(self, **kw):
        msg = kw.get(
            'message',
            'service unavailable',
        )
        response.status = 503
        return dict(message=msg)

    @expose('json')
    def error(self, **kw):
        msg = kw.get(
            'message',
            'an error has occured',
        )
        response.status = 500
        return dict(message=msg)



# -*- coding: utf8 -*-

from ceph_installer import process


class TestMakeAnsibleCommand(object):

    def test_with_tags(self, monkeypatch):
        monkeypatch.setattr(process, 'which', lambda x: "/bin/ansible")
        result = process.make_ansible_command("/hosts", "uuid", tags="package-install")
        assert "--tags" in result
        assert "package-install" in result

    def test_with_skip_tags(self, monkeypatch):
        monkeypatch.setattr(process, 'which', lambda x: "/bin/ansible")
        result = process.make_ansible_command("/hosts", "uuid", skip_tags="package-install")
        assert "--skip-tags" in result
        assert "package-install" in result

    def test_with_extra_vars(self, monkeypatch):
        monkeypatch.setattr(process, 'which', lambda x: "/bin/ansible")
        result = process.make_ansible_command("/hosts", "uuid", extra_vars=dict(foo="bar"))
        assert '{"foo": "bar"}' in result
        assert "--extra-vars" in result

    def test_with_playbook(self, monkeypatch):
        monkeypatch.setattr(process, 'which', lambda x: "/bin/ansible")
        result = process.make_ansible_command("/hosts", "uuid", playbook="playbook.yml")
        assert "/usr/share/ceph-ansible/playbook.yml" in result


class FakePopen(object):

    def __init__(self, stdout='', stderr='', returncode=0):
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode

    def communicate(self, *a):
        return self.stdout, self.stderr


class TestProcess(object):

    def test_decode_unicode_on_the_fly_for_stdout(self, monkeypatch):
        monkeypatch.setattr(
            process.subprocess, 'Popen',
            lambda *a, **kw: FakePopen('£', 'stderr')
        )
        stdout, stderr, code = process.run('ls')
        assert stdout == u'\xa3'

    def test_decode_unicode_on_the_fly_for_stderr(self, monkeypatch):
        monkeypatch.setattr(
            process.subprocess, 'Popen',
            lambda *a, **kw: FakePopen('stdout', '™')
        )
        stdout, stderr, code = process.run('ls')
        assert stderr == u'\u2122'



import logging
import sys
import traceback
from functools import wraps


def catches(catch=None, handler=None, exit=True, handle_all=False, logger=None):
    """
    Very simple decorator that tries any of the exception(s) passed in as
    a single exception class or tuple (containing multiple ones) returning the
    exception message and optionally handling the problem if it raises with the
    handler if it is provided.

    So instead of doing something like this::

        def bar():
            try:
                some_call()
                print "Success!"
            except TypeError, exc:
                print "Error while handling some call: %s" % exc
                sys.exit(1)

    You would need to decorate it like this to have the same effect::

        @catches(TypeError)
        def bar():
            some_call()
            print "Success!"

    If multiple exceptions need to be caught they need to be provided as a
    tuple::

        @catches((TypeError, AttributeError))
        def bar():
            some_call()
            print "Success!"

    If adding a handler, it should accept a single argument, which would be the
    exception that was raised, it would look like::

        def my_handler(exc):
            print 'Handling exception %s' % str(exc)
            raise SystemExit

        @catches(KeyboardInterrupt, handler=my_handler)
        def bar():
            some_call()

    Note that the handler needs to raise its SystemExit if it wants to halt
    execution, otherwise the decorator would continue as a normal try/except
    block.


    :param catch: A tuple with one (or more) Exceptions to catch
    :param handler: Optional handler to have custom handling of exceptions
    :param exit: Raise a ``SystemExit`` after handling exceptions
    :param handle_all: Handle all other exceptions via logging.
    """
    catch = catch or Exception
    logger = logger or logging.getLogger('ceph-installer')

    def decorate(f):

        @wraps(f)
        def newfunc(*a, **kw):
            exit_from_catch = False
            try:
                return f(*a, **kw)
            except catch as e:
                if handler:
                    return handler(e)
                else:
                    logger.error(make_exception_message(e))

                    if exit:
                        exit_from_catch = True
                        sys.exit(1)
            except Exception:  # anything else, no need to save the exception as a variable
                if handle_all is False:  # re-raise if we are not supposed to handle everything
                    raise
                # Make sure we don't spit double tracebacks if we are raising
                # SystemExit from the `except catch` block

                if exit_from_catch:
                    sys.exit(1)

                str_failure = traceback.format_exc()
                for line in str_failure.split('\n'):
                    logger.error("%s" % line)
                sys.exit(1)

        return newfunc

    return decorate

#
# Decorator helpers
#


def make_exception_message(exc):
    """
    An exception is passed in and this function
    returns the proper string depending on the result
    so it is readable enough.
    """
    if str(exc):
        return '%s: %s\n' % (exc.__class__.__name__, exc)
    else:
        return '%s\n' % (exc.__class__.__name__)



# -*- coding: utf8 -*-
from ceph_installer import models, tasks


class TestCallAnsible(object):

    def setup(self):
        models.Task(
            identifier='aaaa',
            endpoint='/api/test/',
        )
        models.commit()

    def test_set_exit_code_on_error(self, session, monkeypatch):
        monkeypatch.setattr(
            tasks.process,
            'make_ansible_command',
            lambda *a, **kw: ['echo']
        )
        tasks.call_ansible.apply(args=([], 'aaaa')).get()
        task = models.Task.get(1)
        assert task.exit_code == -1

    def test_set_exception_on_stderr(self, session, monkeypatch):
        def error(*a, **kw): raise OSError('a severe error')
        monkeypatch.setattr(tasks.process, 'run', error)
        monkeypatch.setattr(
            tasks.process,
            'make_ansible_command',
            lambda *a, **kw: ['echo']
        )
        tasks.call_ansible.apply(args=([], 'aaaa')).get()
        task = models.Task.get(1)
        assert task.stderr == 'a severe error'

    def test_handle_unicode_on_stderr(self, session, monkeypatch):
        monkeypatch.setattr(
            tasks.process, 'run',
            lambda *a, **kw: ['stdout', 'ᓆ'.decode('utf-8', 'replace'), 1])
        monkeypatch.setattr(
            tasks.process,
            'make_ansible_command',
            lambda *a, **kw: ['echo']
        )
        tasks.call_ansible.apply(args=([], 'aaaa')).get()
        task = models.Task.get(1)
        assert task.stderr == u'\u14c6'

    def test_handle_unicode_on_stdout(self, session, monkeypatch):
        monkeypatch.setattr(
            tasks.process, 'run',
            lambda *a, **kw: ['£'.decode('utf-8', 'replace'), 'stderr', 1])
        monkeypatch.setattr(
            tasks.process,
            'make_ansible_command',
            lambda *a, **kw: ['echo']
        )
        tasks.call_ansible.apply(args=([], 'aaaa')).get()
        task = models.Task.get(1)
        assert task.stdout == u'\xa3'

    def test_process_a_request_that_is_none(self, session, monkeypatch):
        monkeypatch.setattr(
            tasks.process,
            'make_ansible_command',
            lambda *a, **kw: ['echo']
        )
        tasks.call_ansible.apply(args=([], 'aaaa')).get()
        task = models.Task.get(1)
        assert task.http_method == ''
        assert task.user_agent == ''
        assert task.request == ''



from pecan import expose, request, response
from webob.static import FileIter
from ceph_installer.util import make_setup_script, make_agent_script
from ceph_installer.controllers import error
import os
from StringIO import StringIO
import logging

logger = logging.getLogger(__name__)


class SetupController(object):

    @expose(content_type='application/octet-stream')
    def index(self):
        script = make_setup_script(request.url)
        response.headers['Content-Disposition'] = 'attachment; filename=setup.sh'
        response.app_iter = FileIter(script)

    @expose(content_type='application/octet-stream')
    def agent(self):
        script = make_agent_script(request.url, request.client_addr)
        response.headers['Content-Disposition'] = 'attachment; filename=agent-setup.sh'
        response.app_iter = FileIter(script)

    @expose(content_type='application/octet-stream')
    def key(self):
        """
        Serves the public SSH key for the user that own the current service
        """
        # look for the ssh key of the current user
        public_key_path = os.path.expanduser('~/.ssh/id_rsa.pub')
        ssh_dir = os.path.dirname(public_key_path)

        if not os.path.isdir(ssh_dir):
            msg = '.ssh directory not found: %s' % ssh_dir
            logger.error(msg)
            error(500, msg)

        if not os.path.exists(public_key_path):
            msg = 'expected public key not found: %s' % public_key_path
            logger.error(msg)
            error(500, msg)

        # define the file to download
        response.headers['Content-Disposition'] = 'attachment; filename=id_rsa.pub'
        with open(public_key_path) as key_contents:
            key = StringIO()
            key.write(key_contents.read())
            key.seek(0)
        response.app_iter = FileIter(key)





class TestSchemaErrors(object):

    def test_install_empty_object(self, session):
        result = session.app.post_json("/api/mon/install/", params=dict(),
                                       expect_errors=True)
        message = result.json['message']
        assert message.endswith('an empty dictionary object was provided')

    def test_install_invalid_json(self, session):
        # note how we are not using post_json for the request
        result = session.app.post("/api/mon/install/", params={2: 1}, expect_errors=True)
        message = result.json['message']
        assert message == 'invalid JSON was received'

    def test_host_is_invalid(self, session):
        params = {'hosts': ''}
        result = session.app.post_json("/api/mon/install/", params=params,
                                       expect_errors=True)
        message = result.json['message']
        assert "hosts" in message
        assert message.endswith(
            "failed validation, requires format: ['host1', 'host2']"
        )

    def test_redhat_storage_is_wrong_type(self, session):
        params = {'hosts': ['node1'], 'redhat_storage': "foo"}
        result = session.app.post_json("/api/mon/install/", params=params,
                                       expect_errors=True)
        message = result.json['message']
        assert message.endswith('not of type boolean')
        assert 'redhat_storage' in message


class TestErrors(object):

    def test_unavailable(self, session):
        result = session.app.get("/errors/unavailable/",
                                 expect_errors=True)
        message = result.json['message']
        assert message == 'service unavailable'

    def test_forbidden(self, session):
        result = session.app.get("/errors/forbidden/",
                                 expect_errors=True)
        message = result.json['message']
        assert message == 'forbidden'

    def test_invalid(self, session):
        result = session.app.get("/errors/forbidden/",
                                 expect_errors=True)
        message = result.json['message']
        assert message == 'forbidden'



import os
import pecan

from celery import Celery
from celery.signals import worker_init
from ceph_installer import models


@worker_init.connect
def bootstrap_pecan(signal, sender):
    try:
        config_path = os.environ['PECAN_CONFIG']
    except KeyError:
        here = os.path.abspath(os.path.dirname(__file__))
        # XXX this will not hold true when installing as a binary
        config_path = os.path.abspath(os.path.join(here, '../config/config.py'))

    pecan.configuration.set_config(config_path, overwrite=True)
    # Once configuration is set we need to initialize the models so that we can connect
    # to the DB wth a configured mapper.
    models.init_model()


app = Celery('ceph_installer.async', broker='amqp://guest@localhost//', include=['ceph_installer.tasks'])
# the default value of CELERYD_CONCURRENCY will be set to the number
# of CPU/cores on the host. This is a problem because we need to ensure
# that there is only one celery worker running.
app.conf.update(CELERYD_CONCURRENCY=1)



from os import path
from pecan import request, redirect

def error(url, msg=None):
    """
    Helper for controller methods to do an internal redirect with
    either a url part like::

        error("/errors/not_allowed/")

    Or an http code::

        error(404)

    The ``msg`` argument is optional and would override the default error
    message for each error condition as defined in the ``ErrorController``
    methods.
    """
    code_to_url = {
        400: '/errors/invalid/',
        403: '/errors/forbidden/',
        404: '/errors/not_found/',
        405: '/errors/not_allowed/',
        500: '/errors/error/',
        503: '/errors/unavailable/',
    }

    if isinstance(url, int):
        url = code_to_url.get(url, 500)

    if msg:
        request.context['message'] = msg
    url = path.join(url, '?message=%s' % msg)
    redirect(url, internal=True)




setup_script = """#!/bin/bash
if [[ $EUID -ne 0 ]]; then
  echo "You must be a root user or execute this script with sudo" 2>&1
  exit 1
fi

if [ ! -f /etc/os-release ]; then
    echo "Unable to determine a supported system"
    echo "will not proceed with installation"
    exit 2
fi

source /etc/os-release
if [ "$ID" != "ubuntu" ] && [ "$ID" != "rhel" ]; then
    echo "Unsupported system detected: $ID"
    echo "will not proceed with installation"
    exit 3
fi


if [ "$ID" == "ubuntu" ]; then
    echo "--> Installing Python 2.7 for Ansible"
    apt-get update
    DEBIAN_FRONTEND=noninteractive apt-get -y install python
fi

echo "--> creating new user with disabled password: ceph-installer"
useradd -m ceph-installer
passwd -d ceph-installer

echo "--> adding provisioning key to the ceph-installer user authorized_keys"
curl -s -L -o ansible.pub {ssh_key_address}
mkdir -m 700 -p /home/ceph-installer/.ssh
cat ansible.pub >> /home/ceph-installer/.ssh/authorized_keys

echo "--> ensuring correct permissions on .ssh/authorized_keys"
chown -R ceph-installer:ceph-installer /home/ceph-installer/.ssh
chmod 600 /home/ceph-installer/.ssh/authorized_keys

echo "--> ensuring that ceph-installer user will be able to sudo"
# write to it wiping everything
echo "ceph-installer ALL=(ALL) NOPASSWD: ALL" > /etc/sudoers.d/ceph-installer

echo "--> ensuring ceph-installer user does not require a tty"
# and now just append
echo 'Defaults:ceph-installer !requiretty' >> /etc/sudoers.d/ceph-installer
"""

# Note that the agent_script can't run by itself, it needs to be concatenated
# along the regular setup script
agent_script = """
echo "--> installing and configuring agent"
curl -d '{{"hosts": ["{target_host}"]}}' -X POST {agent_endpoint}
"""



from pecan import expose


class CalamariController(object):

    @expose('json')
    def index(self):
        # TODO: allow some autodiscovery here so that clients can see what is
        # available
        return dict()

    @expose('json')
    def install(self):
        return {}

    @expose('json')
    def configure(self):
        return {}



from ceph_installer import hooks
from ceph_installer.controllers import status


def generic_system_error():
    msg = "important system is not running"
    raise hooks.SystemCheckError(msg)

error_checks = (
    generic_system_error,
)

ok_checks = (
    lambda: True,
)


class TestSetupController(object):

    def test_index_system_error_message(self, session, monkeypatch):
        monkeypatch.setattr(status, 'system_checks', error_checks)
        result = session.app.get("/api/status/", expect_errors=True)
        assert result.json['message'] == 'important system is not running'

    def test_index_system_error_code(self, session, monkeypatch):
        monkeypatch.setattr(status, 'system_checks', error_checks)
        result = session.app.get("/api/status/", expect_errors=True)
        assert result.status_int == 500

    def test_index_system_ok_message(self, session, monkeypatch):
        monkeypatch.setattr(status, 'system_checks', ok_checks)
        result = session.app.get("/api/status/")
        assert result.json['message'] == "ok"

    def test_index_system_ok_status(self, session, monkeypatch):
        monkeypatch.setattr(status, 'system_checks', ok_checks)
        result = session.app.get("/api/status/")
        assert result.status_int == 200



import os

import pytest
from ceph_installer import util


class TestGenerateInventoryFile(object):

    def test_single_host(self, tmpdir):
        result = util.generate_inventory_file([("mons", "google.com")], "uuid", tmp_dir=str(tmpdir))
        with open(result, 'r') as f:
            data = [line.strip() for line in f.readlines()]
            assert "[mons]" in data
            assert "google.com" in data

    def test_correct_filename(self, tmpdir):
        result = util.generate_inventory_file([("mons", "google.com")], "uuid", tmp_dir=str(tmpdir))
        assert "uuid_" in result

    def test_multiple_hosts(self, tmpdir):
        hosts = ['google.com', 'redhat.com']
        result = util.generate_inventory_file([("mons", hosts)], "uuid", tmp_dir=str(tmpdir))
        with open(result, 'r') as f:
            data = [line.strip() for line in f.readlines()]
            assert "[mons]" in data
            assert "google.com" in data
            assert "redhat.com" in data

    def test_multiple_groups(self, tmpdir):
        hosts = ['google.com', 'redhat.com']
        inventory = [('mons', hosts), ('osds', 'osd1.host')]
        result = util.generate_inventory_file(inventory, "uuid", tmp_dir=str(tmpdir))
        with open(result, 'r') as f:
            data = [line.strip() for line in f.readlines()]
            assert "[mons]" in data
            assert "google.com" in data
            assert "redhat.com" in data
            assert "[osds]" in data
            assert "osd1.host" in data

    def test_tmp_dir(self, tmpdir):
        result = util.generate_inventory_file([("mons", "google.com")], "uuid", tmp_dir=str(tmpdir))
        assert str(tmpdir) in result


class TestGetEndpoint(object):

    def test_no_args(self):
        result = util.get_endpoint('http://example.org/some/endpoint')
        assert result == 'http://example.org/'

    def test_one_arg(self):
        result = util.get_endpoint('http://example.org/some/endpoint', 'setup')
        assert result == 'http://example.org/setup/'

    def test_no_trailing_slash(self):
        result = util.get_endpoint('http://example.org', 'setup')
        assert result == 'http://example.org/setup/'


class TestMkdir(object):

    def test_mkdir_success(self, tmpdir):
        path = os.path.join(str(tmpdir), 'mydir')
        util.mkdir(path)
        assert os.path.isdir(path) is True

    def test_mkdir_ignores_existing_dir(self, tmpdir):
        path = str(tmpdir)
        util.mkdir(path)
        assert os.path.isdir(path) is True

    def test_mkdir_does_not_ignore_existing_dir(self, tmpdir):
        path = str(tmpdir)
        with pytest.raises(OSError):
            util.mkdir(path, exist_ok=False)


class TestGetInstallExtraVars(object):

    def test_no_extra_vars(self):
        data = dict()
        result = util.get_install_extra_vars(data)
        expected = {
            'ceph_stable': True,
            'fetch_directory': os.path.join(os.environ['HOME'], 'fetch'),
        }
        assert result == expected

    def test_redhat_storage_is_true(self):
        data = dict(redhat_storage=True)
        result = util.get_install_extra_vars(data)
        assert "ceph_stable_rh_storage" in result
        assert "ceph_stable_rh_storage_cdn_install" in result

    def test_redhat_storage_is_false(self):
        data = dict(redhat_storage=False)
        result = util.get_install_extra_vars(data)
        expected = {
            'ceph_stable': True,
            'fetch_directory': os.path.join(os.environ['HOME'], 'fetch'),
        }
        assert result == expected

    def test_redhat_use_cdn_is_true(self):
        data = dict(redhat_storage=True, redhat_use_cdn=True)
        result = util.get_install_extra_vars(data)
        assert "ceph_stable_rh_storage" in result
        assert "ceph_stable_rh_storage_cdn_install" in result

    def test_redhat_use_cdn_is_false(self):
        data = dict(redhat_storage=True, redhat_use_cdn=False)
        result = util.get_install_extra_vars(data)
        assert "ceph_stable_rh_storage" in result
        assert "ceph_stable_rh_storage_cdn_install" not in result

    def test_ceph_origin_exists_when_not_using_the_cdn(self):
        data = dict(redhat_storage=True, redhat_use_cdn=False)
        result = util.get_install_extra_vars(data)
        assert "ceph_origin" in result

    def test_ceph_origin_does_not_exists_when_using_the_cdn(self):
        data = dict(redhat_storage=True, redhat_use_cdn=True)
        result = util.get_install_extra_vars(data)
        assert "ceph_origin" not in result


class TestGetOSDConfigureExtraVars(object):

    def setup(self):
        self.data = dict(
            host="node1",
            fsid="1720107309134",
            devices={'/dev/sdb': '/dev/sdc'},
            monitors=[{"host": "mon1.host", "interface": "eth1"}],
            journal_size=100,
            public_network="0.0.0.0/24",
        )

    def test_raw_multi_journal_is_set(self):
        result = util.get_osd_configure_extra_vars(self.data)
        assert "raw_multi_journal" in result

    def test_raw_journal_devices(self):
        result = util.get_osd_configure_extra_vars(self.data)
        assert "raw_journal_devices" in result
        assert result["raw_journal_devices"] == ["/dev/sdc"]

    def test_redhat_storage_not_present(self):
        data = self.data.copy()
        data["redhat_storage"] = True
        result = util.get_osd_configure_extra_vars(self.data)
        assert "redhat_storage" not in result

    def test_devices_should_be_a_list(self):
        # regression
        result = util.get_osd_configure_extra_vars(self.data)
        assert result["devices"] == ["/dev/sdb"]

    def test_monitor_name_is_set(self):
        # simulates the scenario where this host is a mon and an osd
        data = self.data.copy()
        data['host'] = "mon1.host"
        result = util.get_osd_configure_extra_vars(data)
        assert "monitor_name" in result
        assert result['monitor_name'] == "mon1.host"


class TestParseMonitors(object):

    def test_with_interface(self):
        data = [
            {"host": "mon0.host", "interface": "eth0"},
            {"host": "mon1.host", "interface": "eth1"},
        ]
        results = util.parse_monitors(data)
        assert "mon0.host monitor_interface=eth0" in results
        assert "mon1.host monitor_interface=eth1" in results

    def test_with_address(self):
        data = [
            {"host": "mon0.host", "address": "eth0"},
            {"host": "mon1.host", "address": "eth1"},
        ]
        results = util.parse_monitors(data)
        assert "mon0.host monitor_address=eth0" in results
        assert "mon1.host monitor_address=eth1" in results

    def test_one_with_no_interface(self):
        data = [
            {"host": "mon0.host"},
            {"host": "mon1.host", "interface": "eth1"},
        ]
        results = util.parse_monitors(data)
        assert "mon0.host" == results[0]
        assert "mon1.host monitor_interface=eth1" in results

    def test_invalid_extra_var(self):
        data = [
            {"host": "mon1.host", "foo": "bar"},
        ]
        results = util.parse_monitors(data)
        assert "mon1.host" == results[0]


class TestValidateMonitors(object):

    def test_host_given_as_monitor(self):
        data = [
            {"host": "mon1.host", "foo": "bar"},
        ]
        results = util.validate_monitors(data, "mon1.host")
        assert not results

    def test_host_not_given_as_monitor(self):
        data = [
            {"host": "mon1.host", "foo": "bar"},
        ]
        results = util.validate_monitors(data, "mon2.host")
        assert results == data

    def test_host_not_exact_match_in_monitors(self):
        data = [
            {"host": "mon1.host", "foo": "bar"},
        ]
        results = util.validate_monitors(data, "mon1")
        assert results == data

    def test_remove_host_from_monitors_leaving_others(self):
        data = [
            {"host": "mon1.host", "foo": "bar"},
            {"host": "mon2.host", "foo": "bar"},
        ]
        results = util.validate_monitors(data, "mon1.host")
        assert len(results) == 1
        assert results[0]['host'] == "mon2.host"



import os
from pecan import make_app
from ceph_installer import models, process
from ceph_installer.util import mkdir


def ensure_ssh_keys():
    """
    Generate ssh keys as early as possible so that they are available to all
    web server workers immediately when serving ``/setup/key/``. This helper
    does not use logging because it is too early in running the application
    and no logging has been configured yet.
    """
    # look for the ssh key of the current user
    private_key_path = os.path.expanduser('~/.ssh/id_rsa')
    public_key_path = os.path.expanduser('~/.ssh/id_rsa.pub')
    ssh_dir = os.path.dirname(public_key_path)

    if not os.path.isdir(ssh_dir):
        mkdir(ssh_dir)

    # if there isn't one create it
    if not os.path.exists(public_key_path):
        # create one
        command = [
                'ssh-keygen', '-q', '-t', 'rsa',
                '-N', '',
                '-f', private_key_path,
        ]
        out, err, code = process.run(command, send_input='y\n')
        if code != 0:
            raise RuntimeError('ssh-keygen failed: %s %s' % (out, err))


def setup_app(config):
    ensure_ssh_keys()
    models.init_model()
    app_conf = dict(config.app)

    return make_app(
        app_conf.pop('root'),
        logging=getattr(config, 'logging', {}),
        **app_conf
    )



import sys


class colorize(str):
    """
    Pretty simple to use::

        colorize.make('foo').bold
        colorize.make('foo').green
        colorize.make('foo').yellow
        colorize.make('foo').red
        colorize.make('foo').blue

    Otherwise you could go the long way (for example if you are
    testing this class)::

        string = colorize('foo')
        string._set_attributes()
        string.red

    """

    def __init__(self, string):
        self.stdout = sys.__stdout__
        self.appends = ''
        self.prepends = ''
        self.isatty = self.stdout.isatty()

    def _set_attributes(self):
        """
        Sets the attributes here because the str class does not
        allow to pass in anything other than a string to the constructor
        so we can't really mess with the other attributes.
        """
        for k, v in self.__colors__.items():
            setattr(self, k, self.make_color(v))

    def make_color(self, color):
        if not self.isatty or self.is_windows:
            return self
        return color + self + '\033[0m' + self.appends

    @property
    def __colors__(self):
        return  dict(
                blue   = '\033[34m',
                green  = '\033[92m',
                yellow = '\033[33m',
                red    = '\033[91m',
                bold   = '\033[1m',
                ends   = '\033[0m'
        )

    @property
    def is_windows(self):
        if sys.platform == 'win32':
            return True
        return False

    @classmethod
    def make(cls, string):
        """
        A helper method to return itself and workaround the fact that
        the str object doesn't allow extra arguments passed in to the
        constructor
        """
        obj = cls(string)
        obj._set_attributes()
        return obj

#
# Common string manipulations
#
red_arrow = colorize.make('-->').red
blue_arrow = colorize.make('-->').blue
bold_arrow = colorize.make('-->').bold
yellow_arrow = colorize.make('-->').yellow



from ceph_installer.controllers import mon


class TestMonController(object):

    def setup(self):
        self.configure_data = dict(
            monitor_secret="secret",
            public_network="0.0.0.0/24",
            host="node1",
            interface="eth0",
            fsid="1720107309134",
        )

    def test_index_get(self, session):
        result = session.app.get("/api/mon/")
        assert result.status_int == 200

    def test_install_missing_hosts(self, session):
        result = session.app.post_json("/api/mon/install/", params=dict(),
                                       expect_errors=True)
        assert result.status_int == 400

    def test_install_bogus_field(self, session):
        data = dict(hosts=["google.com"], bogus="foo")
        result = session.app.post_json("/api/mon/install/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_install_hosts_not_a_list(self, session):
        data = dict(hosts="google.com")
        result = session.app.post_json("/api/mon/install/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_install_hosts(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = dict(hosts=["node1"])
        result = session.app.post_json("/api/mon/install/", params=data,
                                       expect_errors=True)
        assert result.json['endpoint'] == '/api/mon/install/'
        assert result.json['identifier'] is not None

    def test_configure_missing_fields(self, session):
        data = dict()
        result = session.app.post_json("/api/mon/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_configure_hosts(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        result = session.app.post_json("/api/mon/configure/", params=self.configure_data)
        assert result.json['endpoint'] == '/api/mon/configure/'
        assert result.json['identifier'] is not None

    def test_configure_with_cluster_name(self, session, monkeypatch, argtest):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', argtest)
        self.configure_data['cluster_name'] = "lol"
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        assert "cluster" in extra_vars
        assert "cluster_name" not in extra_vars
        assert extra_vars["cluster"] == "lol"

    def test_configure_with_conf(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        self.configure_data['conf'] = {"global": {"auth supported": "cephx"}}
        result = session.app.post_json("/api/mon/configure/", params=self.configure_data)
        assert result.json['endpoint'] == '/api/mon/configure/'
        assert result.json['identifier'] is not None

    def test_configure_with_invalid_conf(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        self.configure_data['conf'] = {"global": {"auth supported": "cephx"}, "monn": 1}
        result = session.app.post_json("/api/mon/configure/", params=self.configure_data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "unexpected item in data" in result.json["message"]

    def test_configure_monitors_not_a_list(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = self.configure_data.copy()
        data["monitors"] = "invalid"
        result = session.app.post_json("/api/mon/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "monitors" in result.json["message"]

    def test_configure_monitors_not_a_list_of_objects(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = self.configure_data.copy()
        data["monitors"] = "['mon1', 'mon2']"
        result = session.app.post_json("/api/mon/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "monitors" in result.json["message"]

    def test_configure_monitors_missing_host_key(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        mons = [{"foo": "bar"}]
        data = self.configure_data.copy()
        data['monitors'] = mons
        result = session.app.post_json("/api/mon/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "monitors" in result.json["message"]

    def test_configure_monitors_missing_interface_key(self, session, monkeypatch):
        monkeypatch.setattr(mon.call_ansible, 'apply_async', lambda args, kwargs: None)
        mons = [{"host": "mon0.host"}]
        data = self.configure_data.copy()
        data['monitors'] = mons
        result = session.app.post_json("/api/mon/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "monitors" in result.json["message"]

    def test_configure_valid_monitors(self, session, monkeypatch):
        def check(args, kwargs):
            inventory = args[0]
            hosts = inventory[0][1]
            assert "node1 monitor_interface=eth0" in hosts
            assert "mon1.host monitor_interface=eth1" in hosts

        monkeypatch.setattr(mon.call_ansible, 'apply_async', check)
        mons = [{"host": "mon1.host", "interface": "eth1"}]
        data = self.configure_data.copy()
        data["monitors"] = mons
        result = session.app.post_json("/api/mon/configure/", params=data)
        assert result.json['endpoint'] == '/api/mon/configure/'
        assert result.json['identifier'] is not None


class TestMonWithCalamari(object):

    def setup(self):
        self.configure_data = dict(
            calamari=True,
            monitor_secret="secret",
            public_network="0.0.0.0/24",
            host="node1",
            interface="eth0",
            fsid="1720107309134",
        )

    def test_install_with_calamari(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"], "calamari": True}
        session.app.post_json("/api/mon/install/", params=data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        assert extra_vars['calamari'] is True

    def test_install_without_calamari(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"]}
        session.app.post_json("/api/mon/install/", params=data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        assert extra_vars['calamari'] is False

    def test_configure_with_calamari(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        assert extra_vars['calamari'] is True

    def test_configure_without_calamari(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        self.configure_data.pop('calamari')
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        # 'calamari' will not be passed in as it wasn't defined explicitly
        assert extra_vars.get('calamari') is None


class TestMonVerbose(object):

    def setup(self):
        self.configure_data = dict(
            verbose=True,
            monitor_secret="secret",
            public_network="0.0.0.0/24",
            host="node1",
            interface="eth0",
            fsid="1720107309134",
        )
        self.configure_data_addr = dict(
            verbose=True,
            monitor_secret="secret",
            public_network="0.0.0.0/24",
            host="node1",
            address="10.0.0.1",
            fsid="1720107309134",
        )

    def test_install_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"], "verbose": True}
        session.app.post_json("/api/mon/install/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is True

    def test_install_non_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"]}
        session.app.post_json("/api/mon/install/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is False

    def test_configure_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is True

    def test_configure_non_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        self.configure_data.pop('verbose')
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is False

    def test_configure_with_monitor_address(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/mon/configure/", params=self.configure_data_addr)
        args = argtest.kwargs['args'][0][0]
        assert args == ('mons', [u'node1 monitor_address=10.0.0.1'])

    def test_configure_with_monitor_interface(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            mon.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/mon/configure/", params=self.configure_data)
        args = argtest.kwargs['args'][0][0]
        assert args == ('mons', [u'node1 monitor_interface=eth0'])



import logging

from pecan import expose, request
from pecan.ext.notario import validate
from uuid import uuid4

from ceph_installer.controllers import error
from ceph_installer.tasks import call_ansible
from ceph_installer import schemas
from ceph_installer import models

logger = logging.getLogger(__name__)


class RGWController(object):

    @expose('json')
    def index(self):
        # TODO: allow some autodiscovery here so that clients can see what is
        # available
        return dict()

    @expose(generic=True, template='json')
    def install(self):
        error(405)

    @install.when(method='POST', template='json')
    @validate(schemas.install_schema, handler="/errors/schema")
    def install_post(self):
        hosts = request.json.get('hosts')
        identifier = str(uuid4())
        task = models.Task(
            request=request,
            identifier=identifier,
            endpoint=request.path,
        )
        # we need an explicit commit here because the command may finish before
        # we conclude this request
        models.commit()
        kwargs = dict(tags="package-install")
        call_ansible.apply_async(
            args=([('rgws', hosts)], identifier),
            kwargs=kwargs,
        )

        return task

    @expose('json')
    def configure(self):
        return {}



# -*- coding: utf-8 -*-
import subprocess
try:
    from setuptools import setup, find_packages, Command
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages, Command
import re


def read_module_contents():
    with open('ceph_installer/__init__.py') as installer_init:
        return installer_init.read()

module_file = read_module_contents()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))
long_description = open('README.rst').read()
version = metadata['version']


class BumpCommand(Command):
    """ Bump the __version__ number and commit all changes. """

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        version = metadata['version'].split('.')
        version[-1] = str(int(version[-1]) + 1)  # Bump the final part

        try:
            print('old version: %s  new version: %s' %
                  (metadata['version'], '.'.join(version)))
            raw_input('Press enter to confirm, or ctrl-c to exit >')
        except KeyboardInterrupt:
            raise SystemExit("\nNot proceeding")

        old = "__version__ = '%s'" % metadata['version']
        new = "__version__ = '%s'" % '.'.join(version)

        module_file = read_module_contents()
        with open('ceph_installer/__init__.py', 'w') as fileh:
            fileh.write(module_file.replace(old, new))

        # Commit everything with a standard commit message
        cmd = ['git', 'commit', '-a', '-m', 'version %s' % '.'.join(version)]
        print(' '.join(cmd))
        subprocess.check_call(cmd)


class ReleaseCommand(Command):
    """ Tag and push a new release. """

    user_options = [('sign', 's', 'GPG-sign the Git tag and release files')]

    def initialize_options(self):
        self.sign = False

    def finalize_options(self):
        pass

    def run(self):
        # Create Git tag
        tag_name = 'v%s' % version
        cmd = ['git', 'tag', '-a', tag_name, '-m', 'version %s' % version]
        if self.sign:
            cmd.append('-s')
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push Git tag to origin remote
        cmd = ['git', 'push', 'origin', tag_name]
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push package to pypi
        cmd = ['python', 'setup.py', 'sdist', 'upload']
        if self.sign:
            cmd.append('--sign')
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push master to the remote
        cmd = ['git', 'push', 'origin', 'master']
        print(' '.join(cmd))
        subprocess.check_call(cmd)


setup(
    name='ceph-installer',
    version=version,
    description='An HTTP API that provides Ceph installation/configuration endpoints',
    long_description=long_description,
    author='',
    author_email='',
    scripts=['bin/ceph-installer', 'bin/ceph-installer-celery', 'bin/ceph-installer-gunicorn'],
    install_requires=[
        'celery',
        'gunicorn',
        'notario>=0.0.11',
        'pecan>=1',
        'pecan-notario',
        'requests',
        'sqlalchemy',
        'tambo',
    ],
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(exclude=['ez_setup']),
    entry_points="""
        [pecan.command]
        populate=ceph_installer.commands.populate:PopulateCommand
        """,
    cmdclass={'bump': BumpCommand, 'release': ReleaseCommand},
)



from ceph_installer.controllers import osd


class TestOSDController(object):

    def setup(self):
        data = dict(
            host="node1",
            fsid="1720107309134",
            devices={'/dev/sdb': '/dev/sdc'},
            monitors=[{"host": "mon1.host", "address": "10.0.0.1"}],
            journal_size=100,
            public_network="0.0.0.0/24",
        )
        self.configure_data = data

    def test_index_get(self, session):
        result = session.app.get("/api/osd/")
        assert result.status_int == 200

    def test_configure_missing_fields(self, session):
        data = dict()
        result = session.app.post_json("/api/osd/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_configure_get(self, session):
        result = session.app.get("/api/osd/configure", expect_errors=True)
        assert result.status_int == 405

    def test_configure_devices_is_wrong_type(self, session, monkeypatch):
        monkeypatch.setattr(osd.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = self.configure_data.copy()
        data['devices'] = "/dev/sdb"
        result = session.app.post_json("/api/osd/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        message = result.json["message"]
        assert 'devices' in message
        assert 'not of type dictionary' in message

    def test_configure_devices_is_wrong_type_in_key(self, session, monkeypatch):
        monkeypatch.setattr(osd.call_ansible, 'apply_async', lambda args, kwargs: None)
        data = self.configure_data.copy()
        data['devices'] = {"/dev/sdb": 1}
        result = session.app.post_json("/api/osd/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400
        message = result.json["message"]
        assert '/dev/sdb' in message
        assert message.endswith(
                '1 failed validation, not of type string'
        )

    def test_configure_invalid_field(self, session):
        data = self.configure_data.copy()
        data["bogus"] = "invalid"
        result = session.app.post_json("/api/osd/configure/", params=data,
                                       expect_errors=True)
        assert result.status_int == 400

    def test_configure_success(self, session, monkeypatch):
        monkeypatch.setattr(osd.call_ansible, 'apply_async', lambda args, kwargs: None)
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data)
        assert result.status_int == 200
        assert result.json['endpoint'] == '/api/osd/configure/'
        assert result.json['identifier'] is not None

    def test_configure_monitor_hosts_interface(self, session, monkeypatch):
        self.configure_data['monitors'] = [{"host": "mon1.host", "interface": "eth1"}]

        def check(args, kwargs):
            inventory = args[0]
            assert "mon1.host monitor_interface=eth1" in inventory[1][1]

        monkeypatch.setattr(osd.call_ansible, 'apply_async', check)
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data)
        assert result.status_int == 200

    def test_configure_monitor_hosts_address(self, session, monkeypatch):
        def check(args, kwargs):
            inventory = args[0]
            assert "mon1.host monitor_address=10.0.0.1" in inventory[1][1]

        monkeypatch.setattr(osd.call_ansible, 'apply_async', check)
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data)
        assert result.status_int == 200

    def test_configure_redhat_storage(self, session, monkeypatch):
        def check(args, kwargs):
            extra_vars = kwargs["extra_vars"]
            assert "redhat_storage" not in extra_vars
            assert "ceph_stable_rh_storage" in extra_vars

        data = self.configure_data.copy()
        data["redhat_storage"] = True
        monkeypatch.setattr(osd.call_ansible, 'apply_async', check)
        result = session.app.post_json("/api/osd/configure/", params=data)
        assert result.status_int == 200

    def test_configure_playbook(self, session, monkeypatch):
        def check(args, kwargs):
            assert "osd-configure.yml" in kwargs["playbook"]

        monkeypatch.setattr(osd.call_ansible, 'apply_async', check)
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data)
        assert result.status_int == 200

    def test_configure_with_conf(self, session, monkeypatch):
        monkeypatch.setattr(osd.call_ansible, 'apply_async', lambda args, kwargs: None)
        self.configure_data['conf'] = {"global": {"auth supported": "cephx"}}
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data)
        assert result.json['endpoint'] == '/api/osd/configure/'
        assert result.json['identifier'] is not None

    def test_configure_with_invalid_conf(self, session, monkeypatch):
        monkeypatch.setattr(osd.call_ansible, 'apply_async', lambda args, kwargs: None)
        self.configure_data['conf'] = {"global": {"auth supported": "cephx"}, "monn": 1}
        result = session.app.post_json("/api/osd/configure/", params=self.configure_data,
                                       expect_errors=True)
        assert result.status_int == 400
        assert "unexpected item in data" in result.json["message"]

    def test_configure_monitor_name_is_set(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        self.configure_data['host'] = "mon1.host"
        session.app.post_json("/api/osd/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['extra_vars']['monitor_name'] == "mon1.host"

    def test_configure_monitor_name_is_not_set(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/osd/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert "monitor_name" not in kwargs['extra_vars']

    def test_configure_with_cluster_name(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        self.configure_data['cluster_name'] = "lol"
        session.app.post_json("/api/osd/configure/", params=self.configure_data)
        extra_vars = argtest.kwargs['kwargs']['extra_vars']
        assert "cluster" in extra_vars
        assert "cluster_name" not in extra_vars
        assert extra_vars["cluster"] == "lol"


class TestOsdVerbose(object):

    def setup(self):
        self.configure_data = dict(
            verbose=True,
            host="node1",
            fsid="1720107309134",
            devices={'/dev/sdb': '/dev/sdc'},
            monitors=[{"host": "mon1.host", "interface": "eth1"}],
            journal_size=100,
            public_network="0.0.0.0/24",
        )

    def test_install_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"], "verbose": True}
        session.app.post_json("/api/osd/install/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is True

    def test_install_non_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        data = {"hosts": ["node1"]}
        session.app.post_json("/api/osd/install/", params=data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is False

    def test_configure_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        session.app.post_json("/api/osd/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is True

    def test_configure_non_verbose(self, session, monkeypatch, argtest):
        monkeypatch.setattr(
            osd.call_ansible, 'apply_async', argtest)
        self.configure_data.pop('verbose')
        session.app.post_json("/api/osd/configure/", params=self.configure_data)
        kwargs = argtest.kwargs['kwargs']
        assert kwargs['verbose'] is False



from pecan import expose
from ceph_installer.controllers import (
    tasks, mon, osd, rgw, calamari, errors, setup, agent,
    status
)


class ApiController(object):

    @expose('json')
    def index(self):
        # TODO: allow some autodiscovery here so that clients can see what is
        # available
        return dict()

    agent = agent.AgentController()
    tasks = tasks.TasksController()
    mon = mon.MONController()
    osd = osd.OSDController()
    rgw = rgw.RGWController()
    calamari = calamari.CalamariController()
    status = status.StatusController()


class RootController(object):

    @expose('json')
    def index(self):
        return dict()

    api = ApiController()
    errors = errors.ErrorController()
    setup = setup.SetupController()



from pecan.commands.base import BaseCommand
from pecan import conf

from ceph_installer import models


def out(string):
    print "==> %s" % string


class PopulateCommand(BaseCommand):
    """
    Load a pecan environment and initializate the database.
    """

    def run(self, args):
        super(PopulateCommand, self).run(args)
        out("LOADING ENVIRONMENT")
        self.load_app()
        out("BUILDING SCHEMA")
        try:
            out("STARTING A TRANSACTION...")
            models.start()
            models.Base.metadata.create_all(conf.sqlalchemy.engine)
        except:
            models.rollback()
            out("ROLLING BACK... ")
            raise
        else:
            out("COMMITING... ")
            models.commit()



from datetime import datetime
from uuid import uuid4

from ceph_installer.models import Task


class TestTasksController(object):

    def create_task(self, **kw):
        Task(
            identifier=kw.get('identifier', str(uuid4())),
            endpoint='/api/rgw/',
            command='ansible-playbook -i "rgw.example.com," playbook.yml',
            stderr='',
            stdout='',
            started=kw.get('started', datetime.utcnow()),
            ended=kw.get('ended', datetime.utcnow()),
            succeeded=True
            )

    def test_index_get_no_tasks(self, session):
        result = session.app.get("/api/tasks/")
        assert result.json == []

    def test_index_get_single_task(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert len(result.json) == 1

    def test_index_get_single_task_identifier(self, session):
        self.create_task(identifier='uuid-1')
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['identifier'] == 'uuid-1'

    def test_index_get_single_task_endpoint(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['endpoint'] == '/api/rgw/'

    def test_index_get_single_task_command(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['command'] == 'ansible-playbook -i "rgw.example.com," playbook.yml'

    def test_index_get_single_task_stdout(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['stdout'] == ''

    def test_index_get_single_task_stderr(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['stderr'] == ''

    def test_index_get_single_task_started(self, session):
        started = datetime.utcnow()
        self.create_task(started=started)
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['started'] == started.isoformat().replace('T', ' ')

    def test_index_get_single_task_ended(self, session):
        ended = datetime.utcnow()
        self.create_task(ended=ended)
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['ended'] == ended.isoformat().replace('T', ' ')

    def test_index_get_single_task_succeeded(self, session):
        self.create_task()
        session.commit()
        result = session.app.get("/api/tasks/")
        assert result.json[0]['succeeded'] is True


class TestTaskController(object):

    def create_task(self, **kw):
        Task(
            identifier=kw.get('identifier', str(uuid4())),
            endpoint='/api/rgw/',
            command='ansible-playbook -i "rgw.example.com," playbook.yml',
            stderr='',
            stdout='',
            started=kw.get('started', datetime.utcnow()),
            ended=kw.get('ended', datetime.utcnow()),
            succeeded=True
        )

    def test_task_not_found(self, session):
        result = session.app.get(
            '/api/tasks/1234-asdf-1234-asdf/',
            expect_errors=True)
        assert result.status_int == 404

    def test_task_exists_with_metadata(self, session):
        identifier = '1234-asdf-1234-asdf'
        self.create_task(identifier=identifier)
        result = session.app.get('/api/tasks/1234-asdf-1234-asdf/')
        assert result.json['identifier']


class TestTaskControllerRequests(object):

    def create_task(self, **kw):

        Task(
            request=kw.get('request'),
            identifier=kw.get('identifier', '1234-asdf-1234-asdf'),
            endpoint='/api/rgw/',
            command='ansible-playbook -i "rgw.example.com," playbook.yml',
            stderr='',
            stdout='',
            started=kw.get('started', datetime.utcnow()),
            ended=kw.get('ended', datetime.utcnow()),
            succeeded=True
        )

    def test_request_is_none(self, session):
        self.create_task()
        result = session.app.get(
            '/api/tasks/1234-asdf-1234-asdf/',
            expect_errors=True)
        print result.json
        assert result.json['user_agent'] == ''
        assert result.json['http_method'] == ''
        assert result.json['request'] == ''

    def test_request_with_valid_method(self, fake, session):
        fake_request = fake(method='POST')
        self.create_task(request=fake_request)
        result = session.app.get('/api/tasks/1234-asdf-1234-asdf/')
        assert result.json['http_method'] == 'POST'

    def test_request_with_valid_body(self, fake, session):
        fake_request = fake(body='{"host": "example.com"}')
        self.create_task(request=fake_request)
        result = session.app.get('/api/tasks/1234-asdf-1234-asdf/')
        assert result.json['request'] == '{"host": "example.com"}'

    def test_request_with_valid_user_agent(self, fake, session):
        fake_request = fake(user_agent='Mozilla/5.0')
        self.create_task(request=fake_request)
        result = session.app.get('/api/tasks/1234-asdf-1234-asdf/')
        assert result.json['user_agent'] == 'Mozilla/5.0'



from celery.task.control import inspect
from errno import errorcode
from ceph_installer import models
from ceph_installer.util import which
from pecan.hooks import PecanHook
from sqlalchemy.exc import OperationalError
from webob.exc import WSGIHTTPException
import logging


logger = logging.getLogger(__name__)


class SystemCheckError(Exception):

    def __init__(self, message):
        self.message = message


def ansible_exists():
    """
    Perform a simple check to see if ``ansibl-playbook`` executable is present
    in the system where the service is running.
    """
    ansible_path = which('ansible-playbook')
    if not ansible_path:
        raise SystemCheckError('Could not find ansible in system paths')


def celery_has_workers():
    """
    The ``stats()`` call will return different stats/metadata information about
    celery worker(s).  An empty/None result will mean that there aren't any
    celery workers in use.
    """
    stats = inspect().stats()
    if not stats:
        raise SystemCheckError('No running Celery worker was found')


def rabbitmq_is_running():
    """
    If checking for worker stats, an ``IOError`` may be raised depending on the
    problem for the RabbitMQ connection.
    """
    try:
        celery_has_workers()
    except IOError as e:
        msg = "Error connecting to RabbitMQ: " + str(e)
        if len(e.args):
            if errorcode.get(e.args[0]) == 'ECONNREFUSED':
                msg = "RabbitMQ is not running or not reachable"
        raise SystemCheckError(msg)


def database_connection():
    """
    A very simple connection that should succeed if there is a good/correct
    database connection.
    """
    try:
        models.Task.get(1)
    except OperationalError as exc:
        raise SystemCheckError(
            "Could not connect or retrieve information from the database: %s" % exc.message)


system_checks = (
    ansible_exists,
    rabbitmq_is_running,
    celery_has_workers,
    database_connection
)


class CustomErrorHook(PecanHook):
    """
    Ensure a traceback is logged correctly on error conditions.

    When an error condition is detected (one of the callables raises
    a ``SystemCheckError``) it sets the response status to 500 and returns
    a JSON response with the appropriate reason.
    """

    def on_error(self, state, exc):
        if isinstance(exc, WSGIHTTPException):
            if exc.code == 404:
                logger.error("Not Found: %s" % state.request.url)
                return
            # explicit redirect codes that should not be handled at all by this
            # utility
            elif exc.code in [300, 301, 302, 303, 304, 305, 306, 307, 308]:
                return

        logger.exception('unhandled error by ceph-installer')
        for check in system_checks:
            try:
                check()
            except SystemCheckError as system_error:
                state.response.json_body = {'message': system_error.message}
                state.response.status = 500
                state.response.content_type = 'application/json'
                return state.response
